"""
A module containing classes and functions for generating and analyzing linear feedback shift registers.
"""

from __future__ import annotations

from typing import Callable, Type, overload

import numba
import numpy as np
import numpy.typing as npt
from galois import FieldArray, Poly
from galois._domains._function import Function
from galois.typing import ArrayLike, PolyLike
from numba import int64
from typing_extensions import Literal, Self

from .._helper import export, verify_isinstance, verify_literal, verify_only_one_specified, verify_scalar

ADD: Callable[[int, int], int]
SUBTRACT: Callable[[int, int], int]
MULTIPLY: Callable[[int, int], int]
RECIPROCAL: Callable[[int], int]


###############################################################################
# Fibonacci LFSR
###############################################################################


@export
class FLFSR:
    r"""
    Implements a Fibonacci linear-feedback shift register (LFSR).

    Notes:
        A Fibonacci LFSR is defined by its feedback polynomial $f(x)$.

        $$
        f(x) = -c_{0} \cdot x^{n} - c_{1} \cdot x^{n-1} - \dots - c_{n-2} \cdot x^{2} - c_{n-1} \cdot x + 1
        = x^n \cdot c(x^{-1})
        $$

        The feedback polynomial is the reciprocal of the characteristic polynomial $c(x)$ of the linear recurrent
        sequence $y$ produced by the Fibonacci LFSR.

        $$c(x) = x^{n} - c_{n-1} \cdot x^{n-1} - c_{n-2} \cdot x^{n-2} - \dots - c_{1} \cdot x - c_{0}$$

        $$y_t = c_{n-1} \cdot y_{t-1} + c_{n-2} \cdot y_{t-2} + \dots + c_{1} \cdot y_{t-n+2} + c_{0} \cdot y_{t-n+1}$$

        .. code-block:: text
           :caption: Fibonacci LFSR Configuration

                    +--------------@<-------------@<------------@<-------------+
                    |              ^              ^             ^              |
                    |              | c_n-1        | c_n-2       | c_1          | c_0
                    |              | T[0]         | T[1]        | T[n-2]       | T[n-1]
                    |              |              |             |              |
                    v  +--------+  |  +--------+  |             |  +--------+  |
            x[t] -->@->|  S[0]  |--+->|  S[1]  |--+---  ...  ---+->| S[n-1] |--+--> y[t]
                       +--------+     +--------+                   +--------+
                        y[t+n-1]       y[t+n-2]                      y[t+1]

            S[k] = State vector
            T[k] = Taps vector
            x[t] = Input sequence
            y[t] = Output sequence
            @ = Finite field adder

        The shift register taps $T$ are defined left-to-right as $T = [T_0, T_1, \dots, T_{n-2}, T_{n-1}]$.
        The state vector $S$ is also defined left-to-right as $S = [S_0, S_1, \dots, S_{n-2}, S_{n-1}]$.

        In the Fibonacci configuration, the shift register taps are $T = [c_{n-1}, c_{n-2}, \dots, c_1, c_0]$.
        Additionally, the state vector is equal to the next $n$ outputs in reversed order, namely
        $S = [y_{t+n-1}, y_{t+n-2}, \dots, y_{t+2}, y_{t+1}]$.

    References:
        - `Gardner, D. 2019. “Applications of the Galois Model LFSR in Cryptography”.
          <https://hdl.handle.net/2134/21932>`_

    See Also:
        berlekamp_massey

    Examples:
        .. md-tab-set::

            .. md-tab-item:: GF(2)

                Create a Fibonacci LFSR from a degree-4 primitive characteristic polynomial over
                $\mathrm{GF}(2)$.

                .. ipython:: python

                    c = galois.primitive_poly(2, 4); c
                    lfsr = sdr.FLFSR(c)
                    print(lfsr)

                Step the Fibonacci LFSR and produce 10 output symbols.

                .. ipython:: python

                    lfsr.state
                    lfsr.step(10)
                    lfsr.state

            .. md-tab-item:: GF(p)

                Create a Fibonacci LFSR from a degree-4 primitive characteristic polynomial over
                $\mathrm{GF}(7)$.

                .. ipython:: python

                    c = galois.primitive_poly(7, 4); c
                    lfsr = sdr.FLFSR(c)
                    print(lfsr)

                Step the Fibonacci LFSR and produce 10 output symbols.

                .. ipython:: python

                    lfsr.state
                    lfsr.step(10)
                    lfsr.state

            .. md-tab-item:: GF(2^m)

                Create a Fibonacci LFSR from a degree-4 primitive characteristic polynomial over
                $\mathrm{GF}(2^3)$.

                .. ipython:: python

                    c = galois.primitive_poly(2**3, 4); c
                    lfsr = sdr.FLFSR(c)
                    print(lfsr)

                Step the Fibonacci LFSR and produce 10 output symbols.

                .. ipython:: python

                    lfsr.state
                    lfsr.step(10)
                    lfsr.state

            .. md-tab-item:: GF(p^m)

                Create a Fibonacci LFSR from a degree-4 primitive characteristic polynomial over
                $\mathrm{GF}(3^3)$.

                .. ipython:: python

                    c = galois.primitive_poly(3**3, 4); c
                    lfsr = sdr.FLFSR(c)
                    print(lfsr)

                Step the Fibonacci LFSR and produce 10 output symbols.

                .. ipython:: python

                    lfsr.state
                    lfsr.step(10)
                    lfsr.state

    Group:
        sequences-linear-recurrent
    """

    def __init__(
        self,
        characteristic_poly: PolyLike | None = None,
        feedback_poly: PolyLike | None = None,
        state: ArrayLike | None = None,
    ):
        r"""
        Creates a new Fibonacci LFSR.

        Arguments:
            characteristic_poly: The characteristic polynomial
                $c(x) = x^{n} - c_{n-1} \cdot x^{n-1} - c_{n-2} \cdot x^{n-2} - \dots - c_{1} \cdot x - c_{0}$.
            feedback_poly: The feedback polynomial
                $f(x) = -c_{0} \cdot x^{n} - c_{1} \cdot x^{n-1} - \dots - c_{n-2} \cdot x^{2} - c_{n-1} \cdot x + 1$.

                .. note::
                    Either `characteristic_poly` or `feedback_poly` must be specified, but not both.

            state: The initial state vector $S = [S_0, S_1, \dots, S_{n-2}, S_{n-1}]$. The default is `None`
                which corresponds to all ones.

        See Also:
            galois.irreducible_poly, galois.primitive_poly
        """
        verify_only_one_specified(characteristic_poly, feedback_poly)
        if characteristic_poly is not None:
            characteristic_poly = Poly._PolyLike(characteristic_poly)
        elif feedback_poly is not None:
            characteristic_poly = Poly._PolyLike(feedback_poly).reverse()

        verify_isinstance(characteristic_poly, Poly)
        if not characteristic_poly.coeffs[0] == 1:
            raise ValueError(
                f"Argument 'characteristic_poly' must have a n-th degree term of 1, not {characteristic_poly}."
            )
        self._characteristic_poly = characteristic_poly

        if state is None:
            state = self.field.Ones(self.order)
        self._initial_state = self._verify_and_convert_state(state)
        self._state = self.initial_state.copy()

    @classmethod
    def Taps(cls, taps: FieldArray, state: ArrayLike | None = None) -> Self:
        r"""
        Creates a Fibonacci LFSR from its taps.

        Arguments:
            taps: The shift register taps $T = [c_{n-1}, c_{n-2}, \dots, c_1, c_0]$.
            state: The initial state vector $S = [S_0, S_1, \dots, S_{n-2}, S_{n-1}]$. The default is `None`
                which corresponds to all ones.

        Returns:
            A Fibonacci LFSR with taps $T = [c_{n-1}, c_{n-2}, \dots, c_1, c_0]$.

        Examples:
            .. ipython:: python

                c = galois.primitive_poly(7, 4); c
                taps = -c.coeffs[1:]; taps
                lfsr = sdr.FLFSR.Taps(taps)
                print(lfsr)
        """
        verify_isinstance(taps, FieldArray)

        # c(x) = x^{n} - c_{n-1} \cdot x^{n-1} - c_{n-2} \cdot x^{n-2} - \dots - c_{1} \cdot x - c_{0}
        # T = [c_n-1, c_n-2, ..., c_1, c_0]
        coeffs = -taps
        coeffs = np.insert(coeffs, 0, 1)  # Add x^n term
        characteristic_poly = Poly(coeffs)

        return cls(characteristic_poly, state=state)

    def _verify_and_convert_state(self, state: ArrayLike):
        verify_isinstance(state, (tuple, list, np.ndarray, FieldArray))

        state = self.field(state)  # Coerce array-like object to field array

        if not state.size == self.order:
            raise ValueError(
                f"Argument 'state' must have size equal to the degree of the characteristic polynomial, "
                f"not {state.size} and {self.characteristic_poly.degree}."
            )

        return state

    def __repr__(self) -> str:
        """
        A terse representation of the Fibonacci LFSR.

        Examples:
            .. ipython:: python

                c = galois.primitive_poly(7, 4); c
                lfsr = sdr.FLFSR(c)
                lfsr
        """
        return f"<Fibonacci LFSR: c(x) = {self.characteristic_poly} over {self.field.name}>"

    def __str__(self) -> str:
        """
        A formatted string of relevant properties of the Fibonacci LFSR.

        Examples:
            .. ipython:: python

                c = galois.primitive_poly(7, 4); c
                lfsr = sdr.FLFSR(c)
                print(lfsr)
        """
        string = "Fibonacci LFSR:"
        string += f"\n  field: {self.field.name}"
        string += f"\n  characteristic_poly: {self.characteristic_poly}"
        string += f"\n  feedback_poly: {self.feedback_poly}"
        string += f"\n  taps: {self.taps}"
        string += f"\n  order: {self.order}"
        string += f"\n  state: {self.state}"
        string += f"\n  initial_state: {self.initial_state}"

        return string

    def reset(self, state: ArrayLike | None = None):
        r"""
        Resets the Fibonacci LFSR state to the specified state.

        Arguments:
            state: The state vector $S = [S_0, S_1, \dots, S_{n-2}, S_{n-1}]$. The default is `None` which
                corresponds to the initial state.

        Examples:
            Step the Fibonacci LFSR 10 steps to modify its state.

            .. ipython:: python

                c = galois.primitive_poly(7, 4); c
                lfsr = sdr.FLFSR(c); lfsr
                lfsr.state
                lfsr.step(10)
                lfsr.state

            Reset the Fibonacci LFSR state.

            .. ipython:: python

                lfsr.reset()
                lfsr.state

            Create an Fibonacci LFSR and view its initial state.

            .. ipython:: python

                c = galois.primitive_poly(7, 4); c
                lfsr = sdr.FLFSR(c); lfsr
                lfsr.state

            Reset the Fibonacci LFSR state to a new state.

            .. ipython:: python

                lfsr.reset([1, 2, 3, 4])
                lfsr.state
        """
        state = self.initial_state if state is None else state
        self._state = self._verify_and_convert_state(state)

    def __call__(self, x: npt.NDArray[np.int_]) -> FieldArray:
        r"""
        Processes the input symbols $x[n]$ through the Fibonacci LFSR.

        Arguments:
            x: The input symbols $x[n]$.

        Returns:
            The output symbols $y[n]$.
        """
        verify_isinstance(x, np.ndarray)

        y, state = fibonacci_lfsr_step_forward_jit(self.field)(self.taps, self.state, x)

        self._state[:] = state[:]
        if y.size == 1:
            y = y[0]

        return y

    def step(self, steps: int = 1) -> FieldArray:
        """
        Produces the next `steps` output symbols.

        Arguments:
            steps: The direction and number of output symbols to produce. The default is 1. If negative, the
                Fibonacci LFSR will step backwards.

        Returns:
            An array of output symbols of type :obj:`field` with size `abs(steps)`.

        Examples:
            Step the Fibonacci LFSR one output at a time. Notice the first $n$ outputs of a Fibonacci LFSR are
            its state reversed.

            .. ipython:: python

                c = galois.primitive_poly(7, 4)
                lfsr = sdr.FLFSR(c, state=[1, 2, 3, 4]); lfsr
                lfsr.state, lfsr.step()
                lfsr.state, lfsr.step()
                lfsr.state, lfsr.step()
                lfsr.state, lfsr.step()
                lfsr.state, lfsr.step()
                # Ending state
                lfsr.state

            Step the Fibonacci LFSR 5 steps in one call. This is more efficient than iterating one output at a time.

            .. ipython:: python

                c = galois.primitive_poly(7, 4)
                lfsr = sdr.FLFSR(c, state=[1, 2, 3, 4]); lfsr
                lfsr.state
                lfsr.step(5)
                # Ending state
                lfsr.state

            Step the Fibonacci LFSR 5 steps backward. Notice the output sequence is the reverse of the original
            sequence. Also notice the ending state is the same as the initial state.

            .. ipython:: python

                lfsr.step(-5)
                lfsr.state
        """
        verify_scalar(steps, int=True)

        if steps == 0:
            return self.field([])

        x = np.zeros(abs(steps), dtype=self.state.dtype)
        if steps > 0:
            y, state = fibonacci_lfsr_step_forward_jit(self.field)(self.taps, self.state, x)
        else:
            if not self.characteristic_poly.coeffs[-1] > 0:
                raise ValueError(
                    "Can only step the shift register backwards if the c_0 tap is non-zero, "
                    f"not c(x) = {self.characteristic_poly}."
                )
            y, state = fibonacci_lfsr_step_backward_jit(self.field)(self.taps, self.state, x)

        self._state[:] = state[:]
        if y.size == 1:
            y = y[0]

        return y

    def to_galois_lfsr(self) -> GLFSR:
        """
        Converts the Fibonacci LFSR to a Galois LFSR that produces the same output.

        Returns:
            An equivalent Galois LFSR.

        Examples:
            Create a Fibonacci LFSR with a given initial state.

            .. ipython:: python

                c = galois.primitive_poly(7, 4); c
                fibonacci_lfsr = sdr.FLFSR(c, state=[1, 2, 3, 4])
                print(fibonacci_lfsr)

            Convert the Fibonacci LFSR to an equivalent Galois LFSR. Notice the initial state is different.

            .. ipython:: python

                galois_lfsr = fibonacci_lfsr.to_galois_lfsr()
                print(galois_lfsr)

            Step both LFSRs and see that their output sequences are identical.

            .. ipython:: python

                fibonacci_lfsr.step(10)
                galois_lfsr.step(10)
        """
        # See answer to this question https://crypto.stackexchange.com/questions/60634/lfsr-jump-ahead-algorithm

        # The Fibonacci LFSR state F = [f_0, f_1, ..., f_n-1] has the next n outputs of y = [f_n-1, ..., f_1, f_0]
        # This corresponds to the polynomial F(x) = f_n-1*x^n-1 + f_n-2*x^n-2 + ... + f_1*x + f_0
        F = Poly(self.state[::-1])  # Fibonacci output polynomial over GF(q)

        # The Galois LFSR state G = [g_0, g_1, ..., g_n-1] represents the element g_0 + g_1*x + ... + g_n-1*x^n-1
        # in GF(q^n). Let G_i and G_j indicate the state vector at times i and j. The next state G_j = G_i*x % P(x)
        # and y_j = G_i*x // P(x) for j = i + 1. This can be rearranged as G_i*x = y_j*P(x) + G_j. For the
        # Fibonacci LFSR output polynomial F(x), initial Galois LFSR state G_0, finial Galois LFSR state G_n,
        # and characteristic polynomial P(x), the equivalence may be written as G_0*x^n = F(x)*P(x) + G_n or
        # equivalently G_0 = (F(x)*P(x) + G_n) // x^n. The last equation simplifies to G_0 = F(x)*P(x) // x^n because
        # G_n has degree less than n, therefore G_n // x^n = 0.
        P = self.characteristic_poly
        S = F * P // Poly.Identity(self.field) ** self.order
        state = S.coefficients(self.order, order="asc")  # Get coefficients in ascending order

        return GLFSR(self.characteristic_poly, state=state)

    @property
    def field(self) -> Type[FieldArray]:
        """
        The :obj:`~galois.FieldArray` subclass for the finite field that defines the linear arithmetic.

        Examples:
            .. ipython:: python

                c = galois.primitive_poly(7, 4); c
                lfsr = sdr.FLFSR(c); lfsr
                lfsr.field
        """
        return self._characteristic_poly.field

    @property
    def characteristic_poly(self) -> Poly:
        r"""
        The characteristic polynomial $c(x)$ that defines the linear recurrent sequence.

        Notes:
            The characteristic polynomial
            $c(x) = x^{n} - c_{n-1} \cdot x^{n-1} - c_{n-2} \cdot x^{n-2} - \dots - c_{1} \cdot x - c_{0}$
            is the reciprocal of the feedback polynomial $c(x) = x^n f(x^{-1})$.

        Examples:
            .. ipython:: python

                c = galois.primitive_poly(7, 4); c
                lfsr = sdr.FLFSR(c); lfsr
                lfsr.characteristic_poly
                lfsr.characteristic_poly == lfsr.feedback_poly.reverse()

        Group:
            Polynomials

        Order:
            61
        """
        return self._characteristic_poly

    @property
    def feedback_poly(self) -> Poly:
        r"""
        The feedback polynomial $f(x)$ that defines the feedback arithmetic.

        Notes:
            The feedback polynomial
            $f(x) = -c_{0} \cdot x^{n} - c_{1} \cdot x^{n-1} - \dots - c_{n-2} \cdot x^{2} - c_{n-1} \cdot x + 1$
            is the reciprocal of the characteristic polynomial $f(x) = x^n \cdot c(x^{-1})$.

        Examples:
            .. ipython:: python

                c = galois.primitive_poly(7, 4); c
                lfsr = sdr.FLFSR(c); lfsr
                lfsr.feedback_poly
                lfsr.feedback_poly == lfsr.characteristic_poly.reverse()

        Group:
            Polynomials

        Order:
            61
        """
        return self._characteristic_poly.reverse()

    @property
    def taps(self) -> FieldArray:
        r"""
        The shift register taps $T = [c_{n-1}, c_{n-2}, \dots, c_1, c_0]$.

        The taps of the shift register define the linear recurrence relation.

        Examples:
            .. ipython:: python

                c = galois.primitive_poly(7, 4); c
                taps = -c.coeffs[1:]; taps
                lfsr = sdr.FLFSR.Taps(taps); lfsr
                lfsr.taps
        """
        # c(x) = x^{n} - c_{n-1} \cdot x^{n-1} - c_{n-2} \cdot x^{n-2} - \dots - c_{1} \cdot x - c_{0}
        # T = [c_n-1, c_n-2, ..., c_1, c_0]
        return -self._characteristic_poly.coeffs[1:]

    @property
    def order(self) -> int:
        """
        The order of the linear recurrence/linear recurrent sequence.

        The order of a sequence is defined by the degree of the minimal polynomial that produces it.
        """
        return self._characteristic_poly.degree

    @property
    def initial_state(self) -> FieldArray:
        r"""
        The initial state vector $S = [S_0, S_1, \dots, S_{n-2}, S_{n-1}]$.

        Examples:
            .. ipython:: python

                c = galois.primitive_poly(7, 4)
                lfsr = sdr.FLFSR(c, state=[1, 2, 3, 4]); lfsr
                lfsr.initial_state

            The initial state is unaffected as the Fibonacci LFSR is stepped.

            .. ipython:: python

                lfsr.step(10)
                lfsr.initial_state

        Group:
            State

        Order:
            62
        """
        return self._initial_state.copy()

    @property
    def state(self) -> FieldArray:
        r"""
        The current state vector $S = [S_0, S_1, \dots, S_{n-2}, S_{n-1}]$.

        Examples:
            .. ipython:: python

                c = galois.primitive_poly(7, 4)
                lfsr = sdr.FLFSR(c, state=[1, 2, 3, 4]); lfsr
                lfsr.state

            The current state is modified as the Fibonacci LFSR is stepped.

            .. ipython:: python

                lfsr.step(10)
                lfsr.state

        Group:
            State

        Order:
            62
        """
        return self._state.copy()


class fibonacci_lfsr_step_forward_jit(Function):
    """
    Steps the Fibonacci LFSR `steps` forward.

    .. code-block:: text
       :caption: Fibonacci LFSR Configuration

                +--------------@<-------------@<------------@<-------------+
                |              ^              ^             ^              |
                |              | c_n-1        | c_n-2       | c_1          | c_0
                |              | T[0]         | T[1]        | T[n-2]       | T[n-1]
                |              |              |             |              |
                v  +--------+  |  +--------+  |             |  +--------+  |
        x[t] -->@->|  S[0]  |--+->|  S[1]  |--+---  ...  ---+->| S[n-1] |--+--> y[t]
                   +--------+     +--------+                   +--------+
                    y[t+n-1]       y[t+n-2]                      y[t+1]

    Arguments:
        taps: The set of taps T = [c_n-1, c_n-2, ..., c_1, c_0].
        state: The state vector [S_0, S_1, ..., S_n-2, S_n-1]. State will be modified in-place!
        steps: The number of output symbols to produce.
        feedback: `True` indicates to output the feedback value `y_1[t]` (LRS) and `False` indicates to output the
            value out of the shift register `y_2[t]`.

    Returns:
        The output sequence of size `steps`.
    """

    def __call__(self, taps: npt.NDArray, state: npt.NDArray, x: npt.NDArray):
        if self.field.ufunc_mode != "python-calculate":
            state_ = state.astype(np.int64)  # NOTE: This will be modified
            y = self.jit(
                taps.astype(np.int64),
                state_,
                x.astype(np.int64),
            )
            y = y.astype(state.dtype)
        else:
            state_ = state.view(np.ndarray)  # NOTE: This will be modified
            y = self.python(
                taps.view(np.ndarray),
                state_,
                x.view(np.ndarray),
            )
        y = self.field._view(y)

        return y, state_

    def set_globals(self):
        global ADD, MULTIPLY
        ADD = self.field._add.ufunc_call_only
        MULTIPLY = self.field._multiply.ufunc_call_only

    _SIGNATURE = numba.types.FunctionType(int64[:](int64[:], int64[:], int64[:]))

    @staticmethod
    def implementation(taps: npt.NDArray, state: npt.NDArray, x: npt.NDArray):
        nonzero_tap_idxs = np.where(taps != 0)[0]  # The nonzero taps
        y = np.zeros(x.size, dtype=state.dtype)  # The output array

        for i in range(x.size):
            f = x[i]  # The feedback value (0) added to the input
            for j in nonzero_tap_idxs:
                f = ADD(f, MULTIPLY(state[j], taps[j]))

            y[i] = state[-1]  # Output is popped off the shift register
            state[1:] = state[0:-1]  # Shift state rightward
            state[0] = f  # Insert feedback value at leftmost position

        return y


class fibonacci_lfsr_step_backward_jit(Function):
    """
    Steps the Fibonacci LFSR `steps` backward.

    .. code-block:: text
       :caption: Fibonacci LFSR Configuration

                +--------------@<-------------@<------------@<-------------+
                |              ^              ^             ^              |
                |              | c_n-1        | c_n-2       | c_1          | c_0
                |              | T[0]         | T[1]        | T[n-2]       | T[n-1]
                |              |              |             |              |
                v  +--------+  |  +--------+  |             |  +--------+  |
        x[t] -->@->|  S[0]  |--+->|  S[1]  |--+---  ...  ---+->| S[n-1] |--+--> y[t]
                   +--------+     +--------+                   +--------+
                    y[t+n-1]       y[t+n-2]                      y[t+1]

    Arguments:
        taps: The set of taps T = [c_n-1, c_n-2, ..., c_1, c_0].
        state: The state vector [S_0, S_1, ..., S_n-2, S_n-1]. State will be modified in-place!
        steps: The number of output symbols to produce.

    Returns:
        The output sequence of size `steps`.
    """

    def __call__(self, taps: npt.NDArray, state: npt.NDArray, x: npt.NDArray):
        if self.field.ufunc_mode != "python-calculate":
            state_ = state.astype(np.int64)  # NOTE: This will be modified
            y = self.jit(
                taps.astype(np.int64),
                state_,
                x.astype(np.int64),
            )
            y = y.astype(state.dtype)
        else:
            state_ = state.view(np.ndarray)  # NOTE: This will be modified
            y = self.python(
                taps.view(np.ndarray),
                state_,
                x.view(np.ndarray),
            )
        y = self.field._view(y)

        return y, state_

    def set_globals(self):
        global SUBTRACT, MULTIPLY, RECIPROCAL
        SUBTRACT = self.field._subtract.ufunc_call_only
        MULTIPLY = self.field._multiply.ufunc_call_only
        RECIPROCAL = self.field._reciprocal.ufunc_call_only

    _SIGNATURE = numba.types.FunctionType(int64[:](int64[:], int64[:], int64[:]))

    @staticmethod
    def implementation(taps: npt.NDArray, state: npt.NDArray, x: npt.NDArray):
        nonzero_tap_idxs = np.where(taps[:-1] != 0)[0]  # The nonzero taps, except last tap
        y = np.zeros(x.size, dtype=state.dtype)  # The output array

        for i in range(x.size):
            f = SUBTRACT(state[0], x[i])  # The feedback value minus the presumed input
            state[0:-1] = state[1:]  # Shift state leftward

            s = f  # The unknown previous state value
            for j in nonzero_tap_idxs:
                s = SUBTRACT(s, MULTIPLY(state[j], taps[j]))
            s = MULTIPLY(s, RECIPROCAL(taps[-1]))

            y[i] = s  # The previous output was the last value in the shift register
            state[-1] = s  # Assign recovered state to the leftmost position

        return y


###############################################################################
# Galois LFSR
###############################################################################


@export
class GLFSR:
    r"""
    Implements a Galois linear-feedback shift register (LFSR).

    Notes:
        A Galois LFSR is defined by its feedback polynomial $f(x)$.

        $$
        f(x) = -c_{0} \cdot x^{n} - c_{1} \cdot x^{n-1} - \dots - c_{n-2} \cdot x^{2} - c_{n-1} \cdot x + 1
        = x^n \cdot c(x^{-1})
        $$

        The feedback polynomial is the reciprocal of the characteristic polynomial $c(x)$ of the linear recurrent
        sequence $y$ produced by the Galois LFSR.

        $$c(x) = x^{n} - c_{n-1} \cdot x^{n-1} - c_{n-2} \cdot x^{n-2} - \dots - c_{1} \cdot x - c_{0}$$

        $$y_t = c_{n-1} \cdot y_{t-1} + c_{n-2} \cdot y_{t-2} + \dots + c_{1} \cdot y_{t-n+2} + c_{0} \cdot y_{t-n+1}$$

        .. code-block:: text
           :caption: Galois LFSR Configuration

            +--------------+<-------------+<-------------+<-------------+
            |              |              |              |              |
            | c_0          | c_1          | c_2          | c_n-1        |
            | T[0]         | T[1]         | T[2]         | T[n-1]       |
            |              |              |              |              |
            |  +--------+  v  +--------+  v              v  +--------+  |
            +->|  S[0]  |--@->|  S[1]  |--@---  ...   ---@->| S[n-1] |--+--> y[t]
               +--------+     +--------+                    +--------+
                 y[t-1]                                       y[t+1]

            S[k] = State vector
            T[k] = Taps vector
            y[t] = Output sequence
            @ = Finite field adder

        The shift register taps $T$ are defined left-to-right as $T = [T_0, T_1, \dots, T_{n-2}, T_{n-1}]$.
        The state vector $S$ is also defined left-to-right as $S = [S_0, S_1, \dots, S_{n-2}, S_{n-1}]$.

        In the Galois configuration, the shift register taps are $T = [c_0, c_1, \dots, c_{n-2}, c_{n-1}]$.

    References:
        - `Gardner, D. 2019. “Applications of the Galois Model LFSR in Cryptography”.
          <https://hdl.handle.net/2134/21932>`_

    See Also:
        berlekamp_massey

    Examples:
        .. md-tab-set::

            .. md-tab-item:: GF(2)

                Create a Galois LFSR from a degree-4 primitive characteristic polynomial over $\mathrm{GF}(2)$.

                .. ipython:: python

                    c = galois.primitive_poly(2, 4); c
                    lfsr = sdr.GLFSR(c)
                    print(lfsr)

                Step the Galois LFSR and produce 10 output symbols.

                .. ipython:: python

                    lfsr.state
                    lfsr.step(10)
                    lfsr.state

            .. md-tab-item:: GF(p)

                Create a Galois LFSR from a degree-4 primitive characteristic polynomial over
                $\mathrm{GF}(7)$.

                .. ipython:: python

                    c = galois.primitive_poly(7, 4); c
                    lfsr = sdr.GLFSR(c)
                    print(lfsr)

                Step the Galois LFSR and produce 10 output symbols.

                .. ipython:: python

                    lfsr.state
                    lfsr.step(10)
                    lfsr.state

            .. md-tab-item:: GF(2^m)

                Create a Galois LFSR from a degree-4 primitive characteristic polynomial over
                $\mathrm{GF}(2^3)$.

                .. ipython:: python

                    c = galois.primitive_poly(2**3, 4); c
                    lfsr = sdr.GLFSR(c)
                    print(lfsr)

                Step the Galois LFSR and produce 10 output symbols.

                .. ipython:: python

                    lfsr.state
                    lfsr.step(10)
                    lfsr.state

            .. md-tab-item:: GF(p^m)

                Create a Galois LFSR from a degree-4 primitive characteristic polynomial over
                $\mathrm{GF}(3^3)$.

                .. ipython:: python

                    c = galois.primitive_poly(3**3, 4); c
                    lfsr = sdr.GLFSR(c)
                    print(lfsr)

                Step the Galois LFSR and produce 10 output symbols.

                .. ipython:: python

                    lfsr.state
                    lfsr.step(10)
                    lfsr.state

    Group:
        sequences-linear-recurrent
    """

    def __init__(
        self,
        characteristic_poly: PolyLike | None = None,
        feedback_poly: PolyLike | None = None,
        state: ArrayLike | None = None,
    ):
        r"""
        Creates a new Galois LFSR.

        Arguments:
            characteristic_poly: The characteristic polynomial
                $c(x) = x^{n} - c_{n-1} \cdot x^{n-1} - c_{n-2} \cdot x^{n-2} - \dots - c_{1} \cdot x - c_{0}$.
            feedback_poly: The feedback polynomial
                $f(x) = -c_{0} \cdot x^{n} - c_{1} \cdot x^{n-1} - \dots - c_{n-2} \cdot x^{2} - c_{n-1} \cdot x + 1$.

                .. note::
                    Either `characteristic_poly` or `feedback_poly` must be specified, but not both.

            state: The initial state vector $S = [S_0, S_1, \dots, S_{n-2}, S_{n-1}]$. The default is `None`
                which corresponds to all ones.

        See Also:
            galois.irreducible_poly, galois.primitive_poly
        """
        verify_only_one_specified(characteristic_poly, feedback_poly)
        if characteristic_poly is not None:
            characteristic_poly = Poly._PolyLike(characteristic_poly)
        elif feedback_poly is not None:
            characteristic_poly = Poly._PolyLike(feedback_poly).reverse()

        verify_isinstance(characteristic_poly, Poly)
        if not characteristic_poly.coeffs[0] == 1:
            raise ValueError(
                f"Argument 'characteristic_poly' must have a n-th degree term of 1, not {characteristic_poly}."
            )
        self._characteristic_poly = characteristic_poly

        if state is None:
            state = self.field.Ones(self.order)
        self._initial_state = self._verify_and_convert_state(state)
        self._state = self.initial_state.copy()

    @classmethod
    def Taps(cls, taps: FieldArray, state: ArrayLike | None = None) -> Self:
        r"""
        Creates a Galois LFSR from its taps.

        Arguments:
            taps: The shift register taps $T = [c_0, c_1, \dots, c_{n-2}, c_{n-1}]$.
            state: The initial state vector $S = [S_0, S_1, \dots, S_{n-2}, S_{n-1}]$. The default is `None`
                which corresponds to all ones.

        Returns:
            A Galois LFSR with taps $T = [c_0, c_1, \dots, c_{n-2}, c_{n-1}]$.

        Examples:
            .. ipython:: python

                c = galois.primitive_poly(7, 4); c
                taps = -c.coeffs[1:][::-1]; taps
                lfsr = sdr.GLFSR.Taps(taps)
                print(lfsr)
        """
        verify_isinstance(taps, FieldArray)

        # c(x) = x^{n} - c_{n-1} \cdot x^{n-1} - c_{n-2} \cdot x^{n-2} - \dots - c_{1} \cdot x - c_{0}
        # T = [c_0, c_1, ..., c_n-2, c_n-1]
        coeffs = -taps[::-1]
        coeffs = np.insert(coeffs, 0, 1)  # Add x^n term
        characteristic_poly = Poly(coeffs)

        return cls(characteristic_poly, state=state)

    def _verify_and_convert_state(self, state: ArrayLike):
        verify_isinstance(state, (tuple, list, np.ndarray, FieldArray))

        state = self.field(state)  # Coerce array-like object to field array

        # if not state.size == self.order:
        if not state.size == self.order:
            raise ValueError(
                f"Argument 'state' must have size equal to the degree of the characteristic polynomial, "
                f"not {state.size} and {self.characteristic_poly.degree}."
            )

        return state

    def __repr__(self) -> str:
        """
        A terse representation of the Galois LFSR.

        Examples:
            .. ipython:: python

                c = galois.primitive_poly(7, 4); c
                lfsr = sdr.GLFSR(c)
                lfsr
        """
        return f"<Galois LFSR: c(x) = {self.characteristic_poly} over {self.field.name}>"

    def __str__(self) -> str:
        """
        A formatted string of relevant properties of the Galois LFSR.

        Examples:
            .. ipython:: python

                c = galois.primitive_poly(7, 4); c
                lfsr = sdr.GLFSR(c)
                print(lfsr)
        """
        string = "Galois LFSR:"
        string += f"\n  field: {self.field.name}"
        string += f"\n  characteristic_poly: {self.characteristic_poly}"
        string += f"\n  feedback_poly: {self.feedback_poly}"
        string += f"\n  taps: {self.taps}"
        string += f"\n  order: {self.order}"
        string += f"\n  state: {self.state}"
        string += f"\n  initial_state: {self.initial_state}"

        return string

    def reset(self, state: ArrayLike | None = None):
        r"""
        Resets the Galois LFSR state to the specified state.

        Arguments:
            state: The state vector $S = [S_0, S_1, \dots, S_{n-2}, S_{n-1}]$. The default is `None` which
                corresponds to the initial state.

        Examples:
            Step the Galois LFSR 10 steps to modify its state.

            .. ipython:: python

                c = galois.primitive_poly(7, 4); c
                lfsr = sdr.GLFSR(c); lfsr
                lfsr.state
                lfsr.step(10)
                lfsr.state

            Reset the Galois LFSR state.

            .. ipython:: python

                lfsr.reset()
                lfsr.state

            Create an Galois LFSR and view its initial state.

            .. ipython:: python

                c = galois.primitive_poly(7, 4); c
                lfsr = sdr.GLFSR(c); lfsr
                lfsr.state

            Reset the Galois LFSR state to a new state.

            .. ipython:: python

                lfsr.reset([1, 2, 3, 4])
                lfsr.state
        """
        state = self.initial_state if state is None else state
        self._state = self._verify_and_convert_state(state)

    def step(self, steps: int = 1) -> FieldArray:
        """
        Produces the next `steps` output symbols.

        Arguments:
            steps: The direction and number of output symbols to produce. The default is 1. If negative, the
                Galois LFSR will step backwards.

        Returns:
            An array of output symbols of type :obj:`field` with size `abs(steps)`.

        Examples:
            Step the Galois LFSR one output at a time.

            .. ipython:: python

                c = galois.primitive_poly(7, 4)
                lfsr = sdr.GLFSR(c, state=[1, 2, 3, 4]); lfsr
                lfsr.state, lfsr.step()
                lfsr.state, lfsr.step()
                lfsr.state, lfsr.step()
                lfsr.state, lfsr.step()
                lfsr.state, lfsr.step()
                # Ending state
                lfsr.state

            Step the Galois LFSR 5 steps in one call. This is more efficient than iterating one output at a time.

            .. ipython:: python

                c = galois.primitive_poly(7, 4)
                lfsr = sdr.GLFSR(c, state=[1, 2, 3, 4]); lfsr
                lfsr.state
                lfsr.step(5)
                # Ending state
                lfsr.state

            Step the Galois LFSR 5 steps backward. Notice the output sequence is the reverse of the original sequence.
            Also notice the ending state is the same as the initial state.

            .. ipython:: python

                lfsr.step(-5)
                lfsr.state
        """
        verify_scalar(steps, int=True)

        if steps == 0:
            return self.field([])

        if steps > 0:
            y, state = galois_lfsr_step_forward_jit(self.field)(self.taps, self.state, steps)
        else:
            if not self.characteristic_poly.coeffs[-1] > 0:
                raise ValueError(
                    "Can only step the shift register backwards if the c_0 tap is non-zero, "
                    f"not c(x) = {self.characteristic_poly}."
                )
            y, state = galois_lfsr_step_backward_jit(self.field)(self.taps, self.state, abs(steps))

        self._state[:] = state[:]
        if y.size == 1:
            y = y[0]

        return y

    def to_fibonacci_lfsr(self) -> FLFSR:
        """
        Converts the Galois LFSR to a Fibonacci LFSR that produces the same output.

        Returns:
            An equivalent Fibonacci LFSR.

        Examples:
            Create a Galois LFSR with a given initial state.

            .. ipython:: python

                c = galois.primitive_poly(7, 4); c
                galois_lfsr = sdr.GLFSR(c, state=[1, 2, 3, 4])
                print(galois_lfsr)

            Convert the Galois LFSR to an equivalent Fibonacci LFSR. Notice the initial state is different.

            .. ipython:: python

                fibonacci_lfsr = galois_lfsr.to_fibonacci_lfsr()
                print(fibonacci_lfsr)

            Step both LFSRs and see that their output sequences are identical.

            .. ipython:: python

                galois_lfsr.step(10)
                fibonacci_lfsr.step(10)
        """
        output = self.step(self.order)
        state = output[::-1]
        self.step(-self.order)

        # Create a new object so the initial state is set properly
        return FLFSR(self.characteristic_poly, state=state)

    @property
    def field(self) -> Type[FieldArray]:
        """
        The :obj:`~galois.FieldArray` subclass for the finite field that defines the linear arithmetic.

        Examples:
            .. ipython:: python

                c = galois.primitive_poly(7, 4); c
                lfsr = sdr.GLFSR(c); lfsr
                lfsr.field
        """
        return self._characteristic_poly.field

    @property
    def characteristic_poly(self) -> Poly:
        r"""
        The characteristic polynomial $c(x)$ that defines the linear recurrent sequence.

        Notes:
            The characteristic polynomial
            $c(x) = x^{n} - c_{n-1} \cdot x^{n-1} - c_{n-2} \cdot x^{n-2} - \dots - c_{1} \cdot x - c_{0}$
            is the reciprocal of the feedback polynomial $c(x) = x^n f(x^{-1})$.

        Examples:
            .. ipython:: python

                c = galois.primitive_poly(7, 4); c
                lfsr = sdr.GLFSR(c); lfsr
                lfsr.characteristic_poly
                lfsr.characteristic_poly == lfsr.feedback_poly.reverse()

        Group:
            Polynomials

        Order:
            61
        """
        return self._characteristic_poly

    @property
    def feedback_poly(self) -> Poly:
        r"""
        The feedback polynomial $f(x)$ that defines the feedback arithmetic.

        Notes:
            The feedback polynomial
            $f(x) = -c_{0} \cdot x^{n} - c_{1} \cdot x^{n-1} - \dots - c_{n-2} \cdot x^{2} - c_{n-1} \cdot x + 1$
            is the reciprocal of the characteristic polynomial $f(x) = x^n \cdot c(x^{-1})$.

        Examples:
            .. ipython:: python

                c = galois.primitive_poly(7, 4); c
                lfsr = sdr.GLFSR(c); lfsr
                lfsr.feedback_poly
                lfsr.feedback_poly == lfsr.characteristic_poly.reverse()

        Group:
            Polynomials

        Order:
            61
        """
        return self._characteristic_poly.reverse()

    @property
    def taps(self) -> FieldArray:
        r"""
        The shift register taps $T = [c_0, c_1, \dots, c_{n-2}, c_{n-1}]$.

        The taps of the shift register define the linear recurrence relation.

        Examples:
            .. ipython:: python

                c = galois.primitive_poly(7, 4); c
                taps = -c.coeffs[1:][::-1]; taps
                lfsr = sdr.GLFSR.Taps(taps); lfsr
                lfsr.taps
        """
        # c(x) = x^{n} - c_{n-1} \cdot x^{n-1} - c_{n-2} \cdot x^{n-2} - \dots - c_{1} \cdot x - c_{0}
        # T = [c_0, c_1, ..., c_n-2, c_n-1]
        return -self._characteristic_poly.coeffs[1:][::-1]

    @property
    def order(self) -> int:
        """
        The order of the linear recurrence/linear recurrent sequence.

        The order of a sequence is defined by the degree of the minimal polynomial that produces it.
        """
        return self._characteristic_poly.degree

    @property
    def initial_state(self) -> FieldArray:
        r"""
        The initial state vector $S = [S_0, S_1, \dots, S_{n-2}, S_{n-1}]$.

        Examples:
            .. ipython:: python

                c = galois.primitive_poly(7, 4)
                lfsr = sdr.GLFSR(c, state=[1, 2, 3, 4]); lfsr
                lfsr.initial_state

            The initial state is unaffected as the Galois LFSR is stepped.

            .. ipython:: python

                lfsr.step(10)
                lfsr.initial_state

        Group:
            State

        Order:
            62
        """
        return self._initial_state

    @property
    def state(self) -> FieldArray:
        r"""
        The current state vector $S = [S_0, S_1, \dots, S_{n-2}, S_{n-1}]$.

        Examples:
            .. ipython:: python

                c = galois.primitive_poly(7, 4)
                lfsr = sdr.GLFSR(c, state=[1, 2, 3, 4]); lfsr
                lfsr.state

            The current state is modified as the Galois LFSR is stepped.

            .. ipython:: python

                lfsr.step(10)
                lfsr.state

        Group:
            State

        Order:
            62
        """
        return self._state


class galois_lfsr_step_forward_jit(Function):
    """
    Steps the Galois LFSR `steps` forward.

    .. code-block:: text
       :caption: Galois LFSR Configuration

        +--------------+<-------------+<-------------+<-------------+
        |              |              |              |              |
        | c_0          | c_1          | c_2          | c_n-1        |
        | T[0]         | T[1]         | T[2]         | T[n-1]       |
        |  +--------+  v  +--------+  v              v  +--------+  |
        +->|  S[0]  |--+->|  S[1]  |--+---  ...   ---+->| S[n-1] |--+--> y[t]
           +--------+     +--------+                    +--------+
                                                          y[t+1]

    Arguments:
        taps: The set of taps T = [c_0, c_1, ..., c_n-2, c_n-2].
        state: The state vector [S_0, S_1, ..., S_n-2, S_n-1]. State will be modified in-place!
        steps: The number of output symbols to produce.

    Returns:
        The output sequence of size `steps`.
    """

    def __call__(self, taps: npt.NDArray, state: npt.NDArray, steps: int):
        if self.field.ufunc_mode != "python-calculate":
            state_ = state.astype(np.int64)  # NOTE: This will be modified
            y = self.jit(taps.astype(np.int64), state_, steps)
            y = y.astype(state.dtype)
        else:
            state_ = state.view(np.ndarray)  # NOTE: This will be modified
            y = self.python(taps.view(np.ndarray), state_, steps)
        y = self.field._view(y)

        return y, state_

    def set_globals(self):
        global ADD, MULTIPLY
        ADD = self.field._add.ufunc_call_only
        MULTIPLY = self.field._multiply.ufunc_call_only

    _SIGNATURE = numba.types.FunctionType(int64[:](int64[:], int64[:], int64))

    @staticmethod
    def implementation(taps: npt.NDArray, state: npt.NDArray, steps: int):
        n = taps.size
        y = np.zeros(steps, dtype=state.dtype)  # The output array

        for i in range(steps):
            f = state[n - 1]  # The feedback value
            y[i] = f  # The output

            if f == 0:
                state[1:] = state[0:-1]
                state[0] = 0
            else:
                for j in range(n - 1, 0, -1):
                    state[j] = ADD(state[j - 1], MULTIPLY(f, taps[j]))
                state[0] = MULTIPLY(f, taps[0])

        return y


class galois_lfsr_step_backward_jit(Function):
    """
    Steps the Galois LFSR `steps` backward.

    .. code-block:: text
       :caption: Galois LFSR Configuration

        +--------------+<-------------+<-------------+<-------------+
        |              |              |              |              |
        | c_0          | c_1          | c_2          | c_n-1        |
        | T[0]         | T[1]         | T[2]         | T[n-1]       |
        |  +--------+  v  +--------+  v              v  +--------+  |
        +->|  S[0]  |--+->|  S[1]  |--+---  ...   ---+->| S[n-1] |--+--> y[t]
           +--------+     +--------+                    +--------+
                                                          y[t+1]

    Arguments:
        taps: The set of taps T = [c_0, c_1, ..., c_n-2, c_n-2].
        state: The state vector [S_0, S_1, ..., S_n-2, S_n-1]. State will be modified in-place!
        steps: The number of output symbols to produce.

    Returns:
        The output sequence of size `steps`.
    """

    def __call__(self, taps: npt.NDArray, state: npt.NDArray, steps: int):
        if self.field.ufunc_mode != "python-calculate":
            state_ = state.astype(np.int64)  # NOTE: This will be modified
            y = self.jit(taps.astype(np.int64), state_, steps)
            y = y.astype(state.dtype)
        else:
            state_ = state.view(np.ndarray)  # NOTE: This will be modified
            y = self.python(taps.view(np.ndarray), state_, steps)
        y = self.field._view(y)

        return y, state_

    def set_globals(self):
        global SUBTRACT, MULTIPLY, RECIPROCAL
        SUBTRACT = self.field._subtract.ufunc_call_only
        MULTIPLY = self.field._multiply.ufunc_call_only
        RECIPROCAL = self.field._reciprocal.ufunc_call_only

    _SIGNATURE = numba.types.FunctionType(int64[:](int64[:], int64[:], int64))

    @staticmethod
    def implementation(taps: npt.NDArray, state: npt.NDArray, steps: int):
        n = taps.size
        y = np.zeros(steps, dtype=state.dtype)  # The output array

        for i in range(steps):
            f = MULTIPLY(state[0], RECIPROCAL(taps[0]))  # The feedback value

            for j in range(0, n - 1):
                state[j] = SUBTRACT(state[j + 1], MULTIPLY(f, taps[j + 1]))

            state[n - 1] = f
            y[i] = f  # The output

        return y


###############################################################################
# Berlekamp-Massey algorithm
###############################################################################


@overload
def berlekamp_massey(sequence: FieldArray, output: Literal["minimal"] = "minimal") -> Poly: ...


@overload
def berlekamp_massey(sequence: FieldArray, output: Literal["fibonacci"]) -> FLFSR: ...


@overload
def berlekamp_massey(sequence: FieldArray, output: Literal["galois"]) -> GLFSR: ...


@export
def berlekamp_massey(sequence, output="minimal"):
    r"""
    Finds the minimal polynomial $c(x)$ that produces the linear recurrent sequence $y$.

    This function implements the Berlekamp-Massey algorithm.

    Arguments:
        sequence: A linear recurrent sequence $y$ in $\mathrm{GF}(p^m)$.
        output: The output object type.

            - `"minimal"` (default): Returns the minimal polynomial that generates the linear recurrent sequence.
              The minimal polynomial is a characteristic polynomial $c(x)$ of minimal degree.
            - `"fibonacci"`: Returns a Fibonacci LFSR that produces $y$.
            - `"galois"`: Returns a Galois LFSR that produces $y$.

    Returns:
        The minimal polynomial $c(x)$, a Fibonacci LFSR, or a Galois LFSR, depending on the value of `output`.

    Notes:
        The minimal polynomial is the characteristic polynomial $c(x)$ of minimal degree that produces the
        linear recurrent sequence $y$.

        $$c(x) = x^{n} - c_{n-1} \cdot x^{n-1} - c_{n-2} \cdot x^{n-2} - \dots - c_{1} \cdot x - c_{0}$$

        $$y_t = c_{n-1} \cdot y_{t-1} + c_{n-2} \cdot y_{t-2} + \dots + c_{1} \cdot y_{t-n+2} + c_{0} \cdot y_{t-n+1}$$

        For a linear sequence with order $n$, at least $2n$ output symbols are required to determine the
        minimal polynomial.

    References:
        - `Gardner, D. 2019. “Applications of the Galois Model LFSR in Cryptography”.
          <https://hdl.handle.net/2134/21932>`_
        - `Jason Sachs, Linear Feedback Shift Registers for the Uninitiated, Part VI: Sing Along with the
          Berlekamp-Massey Algorithm. <https://www.embeddedrelated.com/showarticle/1099.php>`_
        - `James Massey, Shift-Register Synthesis and BCH Decoding.
          <https://crypto.stanford.edu/~mironov/cs359/massey.pdf>`_

    Examples:
        The sequence below is a degree-4 linear recurrent sequence over $\mathrm{GF}(7)$.

        .. ipython:: python

            GF = galois.GF(7)
            y = GF([5, 5, 1, 3, 1, 4, 6, 6, 5, 5])

        The characteristic polynomial is $c(x) = x^4 + x^2 + 3x + 5$ over $\mathrm{GF}(7)$.

        .. ipython:: python

            sdr.berlekamp_massey(y)

        Use the Berlekamp-Massey algorithm to return equivalent Fibonacci LFSR that reproduces the sequence.

        .. ipython:: python

            lfsr = sdr.berlekamp_massey(y, output="fibonacci")
            print(lfsr)
            z = lfsr.step(y.size); z
            np.array_equal(y, z)

        Use the Berlekamp-Massey algorithm to return equivalent Galois LFSR that reproduces the sequence.

        .. ipython:: python

            lfsr = sdr.berlekamp_massey(y, output="galois")
            print(lfsr)
            z = lfsr.step(y.size); z
            np.array_equal(y, z)

    Group:
        sequences-linear-recurrent
    """
    verify_isinstance(sequence, FieldArray)
    verify_literal(output, ["minimal", "fibonacci", "galois"])
    if not sequence.ndim == 1:  # TODO: Replace this with verify_arraylike() check
        raise ValueError(f"Argument 'sequence' must be 1-D, not {sequence.ndim}-D.")

    field = type(sequence)
    coeffs = berlekamp_massey_jit(field)(sequence)
    characteristic_poly = Poly(coeffs, field=field)

    if output == "minimal":
        return characteristic_poly

    # The first n outputs are the Fibonacci state reversed
    state_ = sequence[0 : characteristic_poly.degree][::-1]
    fibonacci_lfsr = FLFSR(characteristic_poly, state=state_)

    if output == "fibonacci":
        return fibonacci_lfsr
    else:
        return fibonacci_lfsr.to_galois_lfsr()


class berlekamp_massey_jit(Function):
    """
    Finds the minimal polynomial c(x) of the input sequence.
    """

    def __call__(self, sequence):
        if self.field.ufunc_mode != "python-calculate":
            coeffs = self.jit(sequence.astype(np.int64))
            coeffs = coeffs.astype(sequence.dtype)
        else:
            coeffs = self.python(sequence.view(np.ndarray))
        coeffs = self.field._view(coeffs)

        return coeffs

    def set_globals(self):
        global ADD, SUBTRACT, MULTIPLY, RECIPROCAL
        ADD = self.field._add.ufunc_call_only
        SUBTRACT = self.field._subtract.ufunc_call_only
        MULTIPLY = self.field._multiply.ufunc_call_only
        RECIPROCAL = self.field._reciprocal.ufunc_call_only

    _SIGNATURE = numba.types.FunctionType(int64[:](int64[:]))

    @staticmethod
    def implementation(sequence):  # pragma: no cover
        N = sequence.size
        s = sequence
        c = np.zeros(N, dtype=sequence.dtype)
        b = np.zeros(N, dtype=sequence.dtype)
        c[0] = 1  # The polynomial c(x) = 1
        b[0] = 1  # The polynomial b(x) = 1
        L = 0
        m = 1
        bb = 1

        for n in range(0, N):
            d = 0
            for i in range(0, L + 1):
                d = ADD(d, MULTIPLY(s[n - i], c[i]))

            if d == 0:
                m += 1
            elif 2 * L <= n:
                t = c.copy()
                d_bb = MULTIPLY(d, RECIPROCAL(bb))
                for i in range(m, N):
                    c[i] = SUBTRACT(c[i], MULTIPLY(d_bb, b[i - m]))
                L = n + 1 - L
                b = t.copy()
                bb = d
                m = 1
            else:
                d_bb = MULTIPLY(d, RECIPROCAL(bb))
                for i in range(m, N):
                    c[i] = SUBTRACT(c[i], MULTIPLY(d_bb, b[i - m]))
                m += 1

        return c[0 : L + 1]



import numpy as np
import pytest

import sdr


def test_bsc():
    rng = np.random.default_rng()
    p = rng.uniform(0.2, 0.8)
    N = int(1000 / p)
    x = rng.integers(0, 2, N)
    P = [[1 - p, p], [p, 1 - p]]
    dmc = sdr.DiscreteMemorylessChannel(P)
    y = dmc(x)
    assert np.count_nonzero(x != y) / N == pytest.approx(p, rel=1e-1)


def test_bec():
    rng = np.random.default_rng()
    p = rng.uniform(0.2, 0.8)
    N = int(1000 / p)
    x = rng.integers(0, 2, N)
    P = [[1 - p, 0, p], [0, 1 - p, p]]
    dmc = sdr.DiscreteMemorylessChannel(P, Y=[0, 1, -1])
    y = dmc(x)
    assert np.count_nonzero(x != y) / N == pytest.approx(p, rel=1e-1)



"""
A module containing functions to approximate detection performance.
"""

from __future__ import annotations

import numpy as np
import numpy.typing as npt

from .._conversion import db
from .._helper import convert_output, export, verify_arraylike, verify_scalar


@export
def albersheim(p_d: npt.ArrayLike, p_fa: npt.ArrayLike, n_nc: npt.ArrayLike = 1) -> npt.NDArray[np.float64]:
    r"""
    Estimates the minimum input signal-to-noise ratio (SNR) required to achieve the desired probability of detection
    $P_d$.

    Arguments:
        p_d: The desired probability of detection $P_d$ in $(0, 1)$.
        p_fa: The desired probability of false alarm $P_{fa}$ in $(0, 1)$.
        n_nc: The number of non-coherent combinations $N_{nc} \ge 1$.

    Returns:
        The minimum required input SNR $\gamma$ in dB.

    See Also:
        sdr.min_snr

    Notes:
        This function implements Albersheim's equation, given by

        $$A = \ln \frac{0.62}{P_{fa}}$$

        $$B = \ln \frac{P_d}{1 - P_d}$$

        $$
        \text{SNR}_{\text{dB}} =
        -5 \log_{10} N_{nc} + \left(6.2 + \frac{4.54}{\sqrt{N_{nc} + 0.44}}\right)
        \log_{10} \left(A + 0.12AB + 1.7B\right) .
        $$

        The error in the estimated minimum SNR is claimed to be less than 0.2 dB for

        $$0.1 \leq P_d \leq 0.9$$
        $$10^{-7} \leq P_{fa} \leq 10^{-3}$$
        $$1 \le N_{nc} \le 8096 .$$

        Albersheim's equation approximates a linear detector. However, the difference between linear and square-law
        detectors is minimal, so Albersheim's equation finds wide use.

    References:
        - `Mark Richards, Alternative Forms of Albersheim's Equation.
          <https://radarsp.weebly.com/uploads/2/1/4/7/21471216/albersheim_alternative_forms.pdf>`_
        - `Mark Richards, Non-Coherent Gain and its Approximations.
          <https://bpb-us-w2.wpmucdn.com/sites.gatech.edu/dist/5/462/files/2016/12/Noncoherent-Integration-Gain-Approximations.pdf>`_
        - https://www.mathworks.com/help/phased/ref/albersheim.html

    Examples:
        Compare the theoretical minimum required SNR using a linear detector in :func:`sdr.min_snr` with the
        estimated minimum required SNR using Albersheim's approximation in :func:`sdr.albersheim`.

        .. ipython:: python

            p_d = 0.9; \
            p_fa = np.logspace(-12, -1, 21)

            @savefig sdr_albersheim_1.png
            plt.figure(); \
            plt.semilogx(p_fa, sdr.albersheim(p_d, p_fa, n_nc=1), linestyle="--"); \
            plt.semilogx(p_fa, sdr.albersheim(p_d, p_fa, n_nc=2), linestyle="--"); \
            plt.semilogx(p_fa, sdr.albersheim(p_d, p_fa, n_nc=4), linestyle="--"); \
            plt.semilogx(p_fa, sdr.albersheim(p_d, p_fa, n_nc=8), linestyle="--"); \
            plt.semilogx(p_fa, sdr.albersheim(p_d, p_fa, n_nc=16), linestyle="--"); \
            plt.semilogx(p_fa, sdr.albersheim(p_d, p_fa, n_nc=32), linestyle="--"); \
            plt.gca().set_prop_cycle(None); \
            plt.semilogx(p_fa, sdr.min_snr(p_d, p_fa, n_nc=1, detector="linear"), label=1); \
            plt.semilogx(p_fa, sdr.min_snr(p_d, p_fa, n_nc=2, detector="linear"), label=2); \
            plt.semilogx(p_fa, sdr.min_snr(p_d, p_fa, n_nc=4, detector="linear"), label=4); \
            plt.semilogx(p_fa, sdr.min_snr(p_d, p_fa, n_nc=8, detector="linear"), label=8); \
            plt.semilogx(p_fa, sdr.min_snr(p_d, p_fa, n_nc=16, detector="linear"), label=16); \
            plt.semilogx(p_fa, sdr.min_snr(p_d, p_fa, n_nc=32, detector="linear"), label=32); \
            plt.legend(title="$N_{nc}$"); \
            plt.xlabel("Probability of false alarm, $P_{fa}$"); \
            plt.ylabel("Minimum required input SNR (dB)"); \
            plt.title("Minimum required SNR across non-coherent combinations for $P_d = 0.9$\nfrom theory (solid) and Albersheim's approximation (dashed)");

        Compare the theoretical non-coherent gain for a linear detector against the approximation from Albersheim's
        equation. This comparison plots curves for various post-integration probabilities of detection.

        .. ipython:: python

            fig, ax = plt.subplots(1, 2, sharey=True); \
            p_fa = 1e-8; \
            n = np.linspace(1, 100, 31).astype(int);
            for p_d in [0.5, 0.8, 0.95]:
                snr = sdr.min_snr(p_d, p_fa, detector="linear")
                ax[0].semilogx(n, sdr.non_coherent_gain(n, snr, p_fa=p_fa, detector="linear", snr_ref="output"), label=p_d)
            ax[0].semilogx(n, sdr.coherent_gain(n), color="k", label="Coherent"); \
            ax[0].legend(title="$P_d$"); \
            ax[0].set_xlabel("Number of samples, $N_{nc}$"); \
            ax[0].set_ylabel("Non-coherent gain, $G_{nc}$"); \
            ax[0].set_title("Theoretical");
            for p_d in [0.5, 0.8, 0.95]:
                g_nc = sdr.albersheim(p_d, p_fa, 1) - sdr.albersheim(p_d, p_fa, n)
                ax[1].semilogx(n, g_nc, linestyle="--", label=p_d)
            @savefig sdr_albersheim_2.png
            ax[1].semilogx(n, sdr.coherent_gain(n), color="k", label="Coherent"); \
            ax[1].legend(title="$P_d$"); \
            ax[1].set_xlabel("Number of samples, $N_{nc}$"); \
            ax[1].set_ylabel("Non-coherent gain, $G_{nc}$"); \
            ax[1].set_title("Albersheim's approximation");

    Group:
        detection-approximation
    """
    p_d = verify_arraylike(p_d, float=True, inclusive_min=0, inclusive_max=1)
    p_fa = verify_arraylike(p_fa, float=True, inclusive_min=0, inclusive_max=1)
    n_nc = verify_arraylike(n_nc, int=True, positive=True)

    A = np.log(0.62 / p_fa)
    B = np.log(p_d / (1 - p_d))
    snr = -5 * np.log10(n_nc) + (6.2 + (4.54 / np.sqrt(n_nc + 0.44))) * np.log10(A + 0.12 * A * B + 1.7 * B)

    return convert_output(snr)


@export
def peebles(p_d: npt.ArrayLike, p_fa: npt.ArrayLike, n_nc: npt.ArrayLike) -> npt.NDArray[np.float64]:
    r"""
    Estimates the non-coherent integration gain for a given probability of detection $P_d$ and false alarm $P_{fa}$.

    Arguments:
        p_d: The desired probability of detection $P_d$ in $(0, 1)$.
        p_fa: The desired probability of false alarm $P_{fa}$ in $(0, 1)$.
        n_nc: The number of non-coherent combinations $N_{nc} \ge 1$.

    Returns:
        The non-coherent integration gain $G_{nc}$.

    See Also:
        sdr.non_coherent_gain

    Notes:
        This function implements Peebles' equation, given by

        $$
        G_{nc} = 6.79 \cdot \left(1 + 0.253 \cdot P_d\right) \cdot \left(1 + \frac{\log_{10}(1 / P_{fa})}{46.6}\right) \cdot \log_{10}(N_{nc}) \cdot \left(1 - 0.14 \cdot \log_{10}(N_{nc}) + 0.0183 \cdot (\log_{10}(n_{nc}))^2\right)
        $$

        The error in the estimated non-coherent integration gain is claimed to be less than 0.8 dB for

        $$0.5 \leq P_d \leq 0.999$$
        $$10^{-10} \leq P_{fa} \leq 10^{-2}$$
        $$1 \le N_{nc} \le 100 .$$

        Peebles' equation approximates the non-coherent integration gain using a square-law detector.

    References:
        - `Mark Richards, Non-Coherent Gain and its Approximations.
          <https://bpb-us-w2.wpmucdn.com/sites.gatech.edu/dist/5/462/files/2016/12/Noncoherent-Integration-Gain-Approximations.pdf>`_

    Examples:
        Compare the theoretical non-coherent gain for a square-law detector against the approximation from Peebles's
        equation. This comparison plots curves for various post-integration probabilities of detection.

        .. ipython:: python

            fig, ax = plt.subplots(1, 2, sharey=True); \
            p_fa = 1e-8; \
            n = np.linspace(1, 100, 31).astype(int);
            for p_d in [0.5, 0.8, 0.95]:
                snr = sdr.min_snr(p_d, p_fa, detector="square-law")
                ax[0].semilogx(n, sdr.non_coherent_gain(n, snr, p_fa=p_fa, detector="square-law", snr_ref="output"), label=p_d)
            ax[0].semilogx(n, sdr.coherent_gain(n), color="k", label="Coherent"); \
            ax[0].legend(title="$P_d$"); \
            ax[0].set_xlabel("Number of samples, $N_{nc}$"); \
            ax[0].set_ylabel("Non-coherent gain, $G_{nc}$"); \
            ax[0].set_title("Theoretical");
            for p_d in [0.5, 0.8, 0.95]:
                ax[1].semilogx(n, sdr.peebles(p_d, p_fa, n), linestyle="--", label=p_d)
            @savefig sdr_peebles_1.png
            ax[1].semilogx(n, sdr.coherent_gain(n), color="k", label="Coherent"); \
            ax[1].legend(title="$P_d$"); \
            ax[1].set_xlabel("Number of samples, $N_{nc}$"); \
            ax[1].set_ylabel("Non-coherent gain, $G_{nc}$"); \
            ax[1].set_title("Peebles's approximation");

    Group:
        detection-approximation
    """
    p_d = verify_arraylike(p_d, float=True, inclusive_min=0, inclusive_max=1)
    p_fa = verify_arraylike(p_fa, float=True, inclusive_min=0, inclusive_max=1)
    n_nc = verify_arraylike(n_nc, int=True, positive=True)

    g_nc = 6.79  # dB
    g_nc *= 1 + 0.253 * p_d
    g_nc *= 1 + np.log10(1 / p_fa) / 46.6
    g_nc *= np.log10(n_nc)
    g_nc *= 1 - 0.14 * np.log10(n_nc) + 0.0183 * np.log10(n_nc) ** 2

    return convert_output(g_nc)


@export
def shnidman(
    p_d: npt.ArrayLike,
    p_fa: npt.ArrayLike,
    n_nc: npt.ArrayLike = 1,
    swerling: int = 0,
) -> npt.NDArray[np.float64]:
    r"""
    Estimates the minimum input signal-to-noise ratio (SNR) required to achieve the desired probability of detection
    $P_d$ for the Swerling target model.

    Arguments:
        p_d: The desired probability of detection $P_d$ in $(0, 1)$.
        p_fa: The desired probability of false alarm $P_{fa}$ in $(0, 1)$.
        n_nc: The number of non-coherent combinations $N_{nc} \ge 1$.
        swerling: The Swerling target model.

            - 0: Non-fluctuating target.
            - 1: Dwell-to-dwell decorrelation. Rayleigh PDF. Target has many scatterers, none are dominant.
              Set of $N_{nc}$ returned pulses are correlated within a dwell but independent with the next set of
              $N_{nc}$ pulses on the next dwell.
            - 2: Pulse-to-pulse decorrelation. Rayleigh PDF. Target has many scatterers, none are dominant.
              Set of $N_{nc}$ returned pulses are independent from each other within a dwell.
            - 3: Dwell-to-dwell decorrelation. Chi-squared PDF with 4 degrees of freedom. Target has many scatterers,
              with one dominant. Set of $N_{nc}$ returned pulses are correlated within a dwell but independent with the
              next set of $N_{nc}$ pulses on the next dwell.
            - 4: Pulse-to-pulse decorrelation. Chi-squared PDF with 4 degrees of freedom. Target has many scatterers,
              with one dominant. Set of $N_{nc}$ returned pulses are independent from each other within a dwell.
            - 5: Same as Swerling 0.

    Returns:
        The minimum required input SNR $\gamma$ in dB.

    See Also:
        sdr.min_snr

    Notes:
        This function implements Shnidman's equation. The error in the estimated minimum SNR is claimed to be less than
        1 dB for

        $$0.1 \leq P_d \leq 0.99$$
        $$10^{-9} \leq P_{fa} \leq 10^{-3}$$
        $$1 \le N_{nc} \le 100 .$$

    References:
        - `Amalia Barrios, A Methodology for Phased Array Radar Threshold Modeling Using the Advanced Propagation Model (APM).
          <https://apps.dtic.mil/sti/pdfs/AD1040159.pdf>`_
        - https://www.mathworks.com/help/phased/ref/shnidman.html

    Examples:
        Compare the theoretical minimum required SNR across Swerling target models for 1, 3, and 30 non-coherent combinations.

        .. ipython:: python

            p_d = 0.9; \
            p_fa = np.logspace(-12, -1, 21)

            fig, ax = plt.subplots(3, 1, figsize=(8, 12));
            for i, n_nc in enumerate([1, 3, 30]):
                ax[i].semilogx(p_fa, sdr.shnidman(p_d, p_fa, n_nc=n_nc, swerling=0), label=0)
                ax[i].semilogx(p_fa, sdr.shnidman(p_d, p_fa, n_nc=n_nc, swerling=1), label=1)
                ax[i].semilogx(p_fa, sdr.shnidman(p_d, p_fa, n_nc=n_nc, swerling=2), label=2)
                ax[i].semilogx(p_fa, sdr.shnidman(p_d, p_fa, n_nc=n_nc, swerling=3), label=3)
                ax[i].semilogx(p_fa, sdr.shnidman(p_d, p_fa, n_nc=n_nc, swerling=4), label=4)
                ax[i].legend(title="Swerling")
                ax[i].set_xlabel("Probability of false alarm, $P_{fa}$")
                ax[i].set_ylabel("Minimum required input SNR (dB), $\gamma$")
                ax[i].set_title(f"$N_{{nc}} = {n_nc}$")
            @savefig sdr_shnidman_1.png
            fig.suptitle("Minimum required SNR across Swerling target models");

    Group:
        detection-approximation
    """
    p_d = verify_arraylike(p_d, float=True, inclusive_min=0, inclusive_max=1)
    p_fa = verify_arraylike(p_fa, float=True, inclusive_min=0, inclusive_max=1)
    n_nc = verify_arraylike(n_nc, int=True, positive=True)
    verify_scalar(swerling, int=True, inclusive_min=0, inclusive_max=5)

    @np.vectorize
    def _calculate(
        p_d: float,
        p_fa: float,
        n_nc: int,
        swerling: int,
    ) -> float:
        """
        Computes the minimum required input SNR in dB for the Swerling target model.
        """
        snr0 = _shnidman_swerling0(p_d, p_fa, n_nc)

        if swerling in (0, 5):
            K = np.inf
        elif swerling == 1:
            K = 1
        elif swerling == 2:
            K = n_nc
        elif swerling == 3:
            K = 2
        elif swerling == 4:
            K = 2 * n_nc

        C1 = 1 / K * (((17.7006 * p_d - 18.4496) * p_d + 14.5339) * p_d - 3.525)
        C2 = 1 / K * (np.exp(27.31 * p_d - 25.14) + (p_d - 0.8) * (0.7 * np.log(1e-5 / p_fa) + (2 * n_nc - 20) / 80))

        if p_d <= 0.872:
            C = C1
        else:
            C = C1 + C2

        snr = snr0 + C

        return snr

    snr = _calculate(p_d, p_fa, n_nc, swerling)

    return convert_output(snr)


def _shnidman_swerling0(
    p_d: float,
    p_fa: float,
    n_nc: int,
) -> float:
    """
    Computes the minimum required input SNR in dB for a Swerling 0 target model.
    """
    # NOTE: MATLAB uses <= 40, but the original paper uses < 40. Following MATLAB's convention to match the test
    # vectors.
    a = 0 if n_nc <= 40 else 0.25

    eta = np.sqrt(-0.8 * np.log(4 * p_fa * (1 - p_fa)))
    eta += np.sign(p_d - 0.5) * np.sqrt(-0.8 * np.log(4 * (1 - p_d) * p_d))

    snr = eta * (eta + 2 * np.sqrt(n_nc / 2 + (a - 0.25))) / n_nc
    snr = db(snr)

    return snr



"""
A subpackage for various LTI filters.
"""

from ._applications import *
from ._delay import *
from ._design_fir import *
from ._design_multirate import *
from ._fir import *
from ._iir import *
from ._polyphase import *



"""
A module for designing finite impulse response (FIR) multirate filters.
"""

from __future__ import annotations

import math

import numpy as np
import numpy.typing as npt
import scipy.signal

from .._helper import export, verify_arraylike, verify_scalar


@export
def multirate_fir(
    interpolation: int,
    decimation: int = 1,
    polyphase_order: int = 23,
    atten: float = 80,
) -> npt.NDArray[np.float64]:
    r"""
    Designs a multirate FIR filter impulse response $h[n]$ using the Kaiser window method.

    Arguments:
        interpolation: The interpolation rate $P$.
        decimation: The decimation rate $Q$.
        polyphase_order: The order of each polyphase filter. Must be odd, such that the filter lengths are even.
        atten: The stopband attenuation $A_{\text{stop}}$ in dB.

    Returns:
        The multirate filter impulse response $h[n]$.

    References:
        - fred harris, *Multirate Signal Processing for Communication Systems*, Chapter 7: Resampling Filters.
        - https://www.mathworks.com/help/dsp/ref/designmultiratefir.html

    Examples:
        Design a multirate FIR filter for rational resampling by 11/3.

        .. ipython:: python

            h = sdr.multirate_fir(11, 3)

            @savefig sdr_multirate_fir_1.png
            plt.figure(); \
            sdr.plot.impulse_response(h);

            @savefig sdr_multirate_fir_2.png
            plt.figure(); \
            sdr.plot.magnitude_response(h);

    Group:
        dsp-polyphase-filtering
    """
    P = verify_scalar(interpolation, int=True, positive=True)
    Q = verify_scalar(decimation, int=True, positive=True)
    verify_scalar(polyphase_order, int=True, positive=True, odd=True)
    verify_scalar(atten, float=True, positive=True)

    half_length = (polyphase_order + 1) // 2
    B = P if P > 1 else Q  # The number of polyphase branches
    R = max(P, Q)  # Inverse of the filter's fractional bandwidth

    # Compute the filter order, which is length - 1
    N = 2 * half_length * B

    # Compute ideal lowpass filter
    n = np.arange(N + 1)
    h = P / R * np.sinc((n - N // 2) / R)

    # Compute Kaiser window
    # beta = scipy.signal.windows.kaiser_beta(atten)
    if atten >= 50:
        beta = 0.1102 * (atten - 8.71)  # TODO: MATLAB uses 8.71 and SciPy uses 8.7
    elif atten > 21:
        beta = 0.5842 * (atten - 21) ** 0.4 + 0.07886 * (atten - 21)
    else:
        beta = 0
    w = scipy.signal.windows.kaiser(N + 1, beta)

    # Compute windowed filter
    h = h * w

    if not (Q > P > 1 and (half_length * P) % Q != 0):
        # The first and last elements are zero. Remove the last zero so that the filter is evenly
        # partitioned across the polyphase branches. The filter now has length 2 * half_length * B and
        # each polyphase branch has length 2 * half_length.
        h = h[:-1]

    # If the above condition is not true, the first and last elements are non-zero. The filter length
    # is 2 * half_length * B + 1 and each polyphase branch has length 2 * half_length + 1. The final
    # column in the polyphase matrix will be padded with zeros.

    return h


def multirate_fir_linear(rate: int) -> npt.NDArray[np.float64]:
    r"""
    The multirate filter is designed to linearly interpolate between samples.

    The filter coefficients are a length-$2r$ linear ramp $\frac{1}{r} [0, ..., r-1, r, r-1, ..., 1]$.
    The first output sample aligns with the first input sample.
    """
    h = np.zeros(2 * rate, dtype=float)
    h[:rate] = np.arange(0, rate) / rate
    h[rate:] = np.arange(rate, 0, -1) / rate
    return h


def multirate_fir_linear_matlab(rate: int) -> npt.NDArray[np.float64]:
    r"""
    The multirate filter is designed to linearly interpolate between samples.

    The filter coefficients are a length-$2r$ linear ramp $\frac{1}{r} [1, ..., r-1, r, r-1, ..., 0]$.
    This is method MATLAB uses. The first output sample is advanced from the first input sample.
    """
    h = np.zeros(2 * rate, dtype=float)
    h[:rate] = np.arange(1, rate + 1) / rate
    h[rate:] = np.arange(rate - 1, -1, -1) / rate
    return h


def multirate_fir_zoh(rate: int) -> npt.NDArray[np.float64]:
    """
    The multirate filter is designed to be a zero-order hold.

    The filter coefficients are a length-$r$ array of ones.
    """
    h = np.ones(rate, dtype=float)
    return h


@export
def polyphase_decompose(branches: int, taps: npt.ArrayLike) -> npt.NDArray:
    r"""
    Decomposes the prototype filter taps $h[n]$ into the polyphase matrix $h_i[n]$ with $B$ phases.

    Arguments:
        branches: The number of polyphase branches $B$.
        taps: The prototype filter feedforward coefficients $h[n]$.

    Returns:
        The polyphase matrix $h_i[n]$.

    Notes:
        The multirate FIR filter taps $h[n]$ are arranged down the columns of the polyphase matrix
        $h_i[n]$ as follows:

        .. code-block:: text
            :caption: Polyphase Matrix with 3 Phases

            +------+------+------+------+
            | h[0] | h[3] | h[6] | h[9] |
            +------+------+------+------+
            | h[1] | h[4] | h[7] | 0    |
            +------+------+------+------+
            | h[2] | h[5] | h[8] | 0    |
            +------+------+------+------+

    References:
        - fred harris, *Multirate Signal Processing for Communication Systems*, Chapter 7: Resampling Filters.

    Examples:
        Decompose the multirate FIR filter (notional taps for demonstration) into polyphase matrices
        with 3 and 6 phases.

        .. ipython:: python

            h = np.arange(0, 20)
            sdr.polyphase_decompose(3, h)
            sdr.polyphase_decompose(6, h)

    Group:
        dsp-polyphase-filtering
    """
    B = verify_scalar(branches, int=True, positive=True)
    taps = verify_arraylike(taps, ndim=1)

    N = math.ceil(taps.size / B) * B  # Filter length

    # Append zeros to the end of the taps so that the filter length is a multiple of B
    taps_pad = np.pad(taps, (0, N - taps.size), mode="constant")

    # Reshape the taps down the columns of H
    H = taps_pad.reshape(-1, B).T

    return H



import numpy as np
import pytest

import sdr


def test_types():
    bsc = sdr.BinaryErasureChannel(0.5)

    y = bsc(0)
    assert isinstance(y, int)

    y = bsc([0, 1])
    assert isinstance(y, np.ndarray)


def test_bit_erasures():
    rng = np.random.default_rng()
    p = rng.uniform(0.2, 0.8)
    N = int(1000 / p)
    x = rng.integers(0, 2, N)
    bsc = sdr.BinaryErasureChannel(p)
    y = bsc(x)
    assert np.count_nonzero(x != y) / N == pytest.approx(p, rel=1e-1)


def test_capacities_exceptions():
    with pytest.raises(ValueError):
        # Erasure probability must be between 0 and 1
        sdr.BinaryErasureChannel.capacities(-0.1)
    with pytest.raises(ValueError):
        # Erasure probability must be between 0 and 1
        sdr.BinaryErasureChannel.capacities(1.1)


def test_capacities_limit_cases():
    assert sdr.BinaryErasureChannel.capacities(0) == 1
    assert sdr.BinaryErasureChannel.capacities(1) == 0


def test_capacities_outputs():
    p = [0, 0.5, 1]
    C = sdr.BinaryErasureChannel.capacities(p)
    assert isinstance(C, np.ndarray)
    assert np.array_equal(C, [1, 0.5, 0])

    C = sdr.BinaryErasureChannel.capacities(0.5)
    assert isinstance(C, float)
    assert C == 0.5



import numpy as np
import pytest

import sdr


@pytest.mark.parametrize("n_nc", [1, 3, 71])
@pytest.mark.parametrize("n_c", [1, 7])
@pytest.mark.parametrize("complex", [True, False])
@pytest.mark.parametrize("detector", ["coherent", "linear", "square-law"])
def test_pdfs(detector, complex, n_c, n_nc):
    rng = np.random.default_rng()

    snr = rng.uniform(-10, 10)  # dB
    sigma2 = 1
    A2 = sdr.linear(snr) * sigma2

    if detector == "coherent":
        # Non-coherent integration is not possible when using a coherent detector
        n_nc = None

    # Compute the theoretical distributions
    h0 = sdr.h0(sigma2, detector=detector, complex=complex, n_c=n_c, n_nc=n_nc)
    h1 = sdr.h1(snr, sigma2, detector=detector, complex=complex, n_c=n_c, n_nc=n_nc)

    # Given a false rate, determine the detection threshold and theoretical probability of detection
    p_fa = 1e-1
    threshold = sdr.threshold(p_fa, sigma2, detector=detector, complex=complex, n_c=n_c, n_nc=n_nc)
    p_d = sdr.p_d(snr, p_fa, detector=detector, complex=complex, n_c=n_c, n_nc=n_nc)

    # Simulate the input signal x for the h0 and h1 hypotheses
    trials = 10_000 * n_c * (1 if n_nc is None else n_nc)
    if complex:
        x_h0 = rng.normal(size=trials, scale=np.sqrt(sigma2 / 2)) + 1j * rng.normal(
            size=trials, scale=np.sqrt(sigma2 / 2)
        )
    else:
        x_h0 = rng.normal(size=trials, scale=np.sqrt(sigma2))
    x_h1 = np.sqrt(A2) + x_h0

    # Perform coherent integration
    x_h0 = np.sum(x_h0.reshape((-1, n_c)), axis=-1)
    x_h1 = np.sum(x_h1.reshape((-1, n_c)), axis=-1)

    # Apply the detection scheme
    if detector == "coherent":
        z_h0 = np.real(x_h0)
        z_h1 = np.real(x_h1)
    elif detector == "linear":
        z_h0 = np.abs(x_h0)
        z_h1 = np.abs(x_h1)
    elif detector == "square-law":
        z_h0 = np.abs(x_h0) ** 2
        z_h1 = np.abs(x_h1) ** 2

    # Perform non-coherent integration
    if n_nc is not None and n_nc > 1:
        z_h0 = np.sum(z_h0.reshape((-1, n_nc)), axis=-1)
        z_h1 = np.sum(z_h1.reshape((-1, n_nc)), axis=-1)

    # Measure probability of false alarm and detection
    p_fa_meas = np.mean(z_h0 > threshold)
    p_d_meas = np.mean(z_h1 > threshold)

    try:
        if h0.mean() == 0:
            assert np.mean(z_h0) == pytest.approx(h0.mean(), abs=0.1)
        else:
            assert np.mean(z_h0) == pytest.approx(h0.mean(), rel=0.2)
        assert np.var(z_h0) == pytest.approx(h0.var(), rel=0.2)

        assert np.mean(z_h1) == pytest.approx(h1.mean(), rel=0.2)
        assert np.var(z_h1) == pytest.approx(h1.var(), rel=0.2)

        assert p_fa_meas == pytest.approx(p_fa, rel=0.2)
        assert p_d_meas == pytest.approx(p_d, rel=0.2)
    except AssertionError as e:
        # import matplotlib.pyplot as plt

        # plt.figure()
        # sdr.plot.detector_pdfs(h0, h1, threshold, p_h0=1e-6, p_h1=1e-3)
        # plt.gca().set_prop_cycle(None)
        # plt.hist(z_h0, bins=101, histtype="step", density=True, label="H0")
        # plt.hist(z_h1, bins=101, histtype="step", density=True, label="H1")
        # plt.legend()
        # plt.suptitle(detector)
        # plt.show()

        raise e



"""
A module containing a class for measuring bit error rates (BER) or symbol error rates (SER).
"""

from __future__ import annotations

import numpy as np
import numpy.typing as npt

from .._helper import export, verify_arraylike, verify_same_shape, verify_scalar


@export
class ErrorRate:
    r"""
    A class for measuring bit error rates (BER) or symbol error rates (SER).

    See Also:
        sdr.pack, sdr.unpack, sdr.hexdump

    .. _error-rate-example:

    Examples:
        Create a new bit error rate measurement object.

        .. ipython:: python

            ber = sdr.ErrorRate()
            x = [1, 1, 1, 1, 1]  # Reference bit vector

        Measure and accumulate bit errors from the first received bit vector containing 1 bit error at 10 dB SNR.
        The bit error rate of the first received bit vector is 0.2.

        .. ipython:: python

            x_hat = [1, 0, 1, 1, 1]
            ber.add(10, x, x_hat)

        Measure and accumulate bit errors from the second received bit vector containing 2 bit errors at 10 dB SNR.
        The bit error rate of the second received bit vector is 0.4.

        .. ipython:: python

            x_hat = [1, 0, 1, 0, 1]
            ber.add(10, x, x_hat)

        The total errors are 3, total bits 10, and average bit error rate 0.3.

        .. ipython:: python

            ber.errors(10), ber.counts(10), ber.error_rate(10)

        Average bit error rates for every SNR can be obtained as follows.

        .. ipython:: python

            ber.error_rates()

    Group:
        measurement-modulation
    """

    def __init__(self) -> None:
        """
        Creates a new error rate tabulation object.
        """
        self._errors: dict[float, int] = {}
        self._counts: dict[float, int] = {}

    def add(
        self,
        snr: float,
        x: npt.ArrayLike,
        x_hat: npt.ArrayLike,
    ) -> tuple[int, int, int]:
        r"""
        Measures the number of bit or symbol errors at the given signal-to-noise ratio (SNR).

        Arguments:
            snr: The signal-to-noise ratio (SNR) in dB. This can be $E_b/N_0$, $E_s/N_0$, $S/N$, $C/N_0$,
                or other SNR quantities. However, users are cautioned to be consistent for a given class instance.
            x: The transmitted bits or symbols $x[k]$.
            x_hat: The received bits or symbols $\hat{x}[k]$.

        Returns:
            - The number of errors.
            - The number of bits or symbols.
            - The error rate.

        See Also:
            sdr.pack, sdr.unpack

        Examples:
            See the class :ref:`error-rate-example` section.
        """
        verify_scalar(snr, float=True)
        x = verify_arraylike(x)
        x_hat = verify_arraylike(x_hat)
        verify_same_shape(x, x_hat)

        errors = np.sum(x_hat != x)
        counts = x.size
        error_rate = errors / counts

        self._errors[snr] = self._errors.get(snr, 0) + errors
        self._counts[snr] = self._counts.get(snr, 0) + counts

        return errors, counts, error_rate

    def errors(self, snr: float) -> int:
        """
        Returns the number of errors at the specified signal-to-noise ratio (SNR).

        Arguments:
            snr: The signal-to-noise ratio (SNR) in dB. This can be $E_b/N_0$, $E_s/N_0$, $S/N$, $C/N_0$,
                or other SNR quantities. However, users are cautioned to be consistent for a given class instance.

        Returns:
            The number of errors at the specified SNR.

        Examples:
            See the class :ref:`error-rate-example` section.
        """
        return self._errors.get(snr, 0)

    def counts(self, snr: float) -> int:
        """
        Returns the number of counts at the specified signal-to-noise ratio (SNR).

        Arguments:
            snr: The signal-to-noise ratio (SNR) in dB. This can be $E_b/N_0$, $E_s/N_0$, $S/N$, $C/N_0$,
                or other SNR quantities. However, users are cautioned to be consistent for a given class instance.

        Returns:
            The number of counts at the specified SNR.

        Examples:
            See the class :ref:`error-rate-example` section.
        """
        return self._counts.get(snr, 0)

    def error_rate(self, snr: float) -> float:
        """
        Returns the error rate at the specified signal-to-noise ratio (SNR).

        Arguments:
            snr: The signal-to-noise ratio (SNR) in dB. This can be $E_b/N_0$, $E_s/N_0$, $S/N$, $C/N_0$,
                or other SNR quantities. However, users are cautioned to be consistent for a given class instance.

        Returns:
            The error rate at the specified SNR.

        Examples:
            See the class :ref:`error-rate-example` section.
        """
        return self.errors(snr) / self.counts(snr)

    def error_rates(self) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
        r"""
        Returns all signal-to-noise ratios (SNRs) in ascending order and their corresponding error rates.

        Returns:
            - The signal-to-noise ratios (SNRs). The specific SNR quantity (e.g., $E_b/N_0$, $E_s/N_0$, $S/N$,
              $C/N_0$) is whatever was consistently provided to :meth:`add`.
            - The bit or symbol error rates.

        Examples:
            See the class :ref:`error-rate-example` section.
        """
        snrs = np.array(sorted(self._errors.keys()))
        error_rates = np.array([self.error_rate(snr) for snr in snrs])
        return snrs, error_rates



"""
A module for various filter helper functions.
"""

from __future__ import annotations

from typing import Any

import numpy as np
import numpy.typing as npt
import scipy.signal


def get_frequency_vector(
    freqs: int | float | npt.ArrayLike = 1024,
    sample_rate: float = 1.0,
    whole: bool = True,
    decades: int | None = None,
) -> npt.NDArray[np.float64]:
    if isinstance(freqs, int):
        # freqs represents the number of frequency points
        if whole:
            max_f = sample_rate
        else:
            max_f = sample_rate / 2

        if decades is None:
            f = np.linspace(0, max_f, freqs, endpoint=False)
        elif isinstance(decades, int):
            f = np.logspace(np.log10(max_f) - decades, np.log10(max_f), freqs, endpoint=False)
        else:
            raise TypeError(f"Argument 'decades' must be an integer, not {type(decades).__name__}.")
    else:
        # freqs represents the single frequency or multiple frequencies
        f = np.asarray(freqs, dtype=float)
        f = np.atleast_1d(f)
        if not f.ndim <= 1:
            raise ValueError(f"Argument 'freqs' must be 0-D or 1-D, not {f.ndim}-D.")

    return f


def frequency_response(
    b: npt.ArrayLike,
    a: npt.ArrayLike,
    freqs: int | float | npt.ArrayLike = 1024,
    sample_rate: float = 1.0,
    whole: bool = True,
    decades: int | None = None,
) -> Any:
    f = get_frequency_vector(freqs, sample_rate, whole, decades)
    f, H = scipy.signal.freqz(b, a, worN=f, fs=sample_rate)
    return f, H



import numpy as np

import sdr


def test_bpsk_rect():
    rng = np.random.default_rng()
    phi = rng.uniform(0, 360)
    sps = 10
    h = sdr.rectangular(sps, norm="power")
    psk = sdr.PSK(2, phase_offset=phi, sps=sps, pulse_shape=h)
    _verify_modulation(psk)


def test_bpsk_rc():
    rng = np.random.default_rng()
    phi = rng.uniform(0, 360)
    sps = 10
    h = sdr.raised_cosine(0.5, 16, sps, norm="power")
    psk = sdr.PSK(2, phase_offset=phi, sps=sps, pulse_shape=h)
    _verify_modulation(psk)


def test_qpsk_rect():
    rng = np.random.default_rng()
    phi = rng.uniform(0, 360)
    sps = 10
    h = sdr.rectangular(sps, norm="power")
    psk = sdr.PSK(4, phase_offset=phi, sps=sps, pulse_shape=h)
    _verify_modulation(psk)


def test_qpsk_rc():
    rng = np.random.default_rng()
    phi = rng.uniform(0, 360)
    sps = 10
    h = sdr.raised_cosine(0.5, 16, sps, norm="power")
    psk = sdr.PSK(4, phase_offset=phi, sps=sps, pulse_shape=h)
    _verify_modulation(psk)


def test_8psk_rect():
    rng = np.random.default_rng()
    phi = rng.uniform(0, 360)
    sps = 10
    h = sdr.rectangular(sps, norm="power")
    psk = sdr.PSK(8, phase_offset=phi, sps=sps, pulse_shape=h)
    _verify_modulation(psk)


def test_8psk_rc():
    rng = np.random.default_rng()
    phi = rng.uniform(0, 360)
    sps = 10
    h = sdr.raised_cosine(0.5, 16, sps, norm="power")
    psk = sdr.PSK(8, phase_offset=phi, sps=sps, pulse_shape=h)
    _verify_modulation(psk)


def _verify_modulation(psk: sdr.PSK):
    rng = np.random.default_rng()
    s = rng.integers(0, psk.order, 20)
    a = psk.map_symbols(s)
    x = psk.modulate(s)

    offset = psk.pulse_shape.size // 2
    a_hat = x[offset : offset + s.size * psk.sps : psk.sps]

    # import matplotlib.pyplot as plt

    # plt.figure()
    # sdr.plot.constellation(a)
    # sdr.plot.constellation(a_hat)

    # plt.figure()
    # sdr.plot.time_domain(a)
    # sdr.plot.time_domain(a_hat)
    # plt.show()

    np.testing.assert_array_almost_equal(a, a_hat, decimal=3)



"""
A module containing functions related to coherent integration.
"""

from __future__ import annotations

import numpy as np
import numpy.typing as npt
import scipy.optimize

from .._conversion import db
from .._helper import convert_output, export, verify_arraylike


@export
def coherent_gain(time_bandwidth: npt.ArrayLike) -> npt.NDArray[np.float64]:
    r"""
    Computes the SNR improvement by coherent integration.

    Arguments:
        time_bandwidth: The time-bandwidth product $T_c B_n$ in seconds-Hz (unitless). If the noise bandwidth equals
            the sample rate, the argument equals the number of samples $N_c$ to coherently integrate.

    Returns:
        The coherent gain $G_c$ in dB.

    Notes:
        The signal $x[n]$ is coherently integrated over $N_c$ samples to produce the output $y[n]$.

        $$y[n] = \sum_{m=0}^{N_c-1} x[n-m]$$

        The coherent integration gain is the reduction in SNR of $x[n]$ compared to $y[n]$, such that both signals
        have the same detection performance.

        $$\text{SNR}_{y,\text{dB}} = \text{SNR}_{x,\text{dB}} + G_c$$

        The coherent integration gain is the time-bandwidth product

        $$G_c = 10 \log_{10} (T_c B_n) .$$

        If the noise bandwidth equals the sample rate, the coherent gain is simply

        $$G_c = 10 \log_{10} N_c .$$

    Examples:
        See the :ref:`coherent-integration` example.

        Compute the coherent gain for various integration lengths.

        .. ipython:: python

            sdr.coherent_gain(1)
            sdr.coherent_gain(2)
            sdr.coherent_gain(10)
            sdr.coherent_gain(20)

        Plot coherent gain as a function of the number of coherently integrated samples.

        .. ipython:: python

            n_c = np.logspace(0, 3, 1001)

            @savefig sdr_coherent_gain_1.png
            plt.figure(); \
            plt.semilogx(n_c, sdr.coherent_gain(n_c)); \
            plt.xlabel("Number of samples, $N_c$"); \
            plt.ylabel("Coherent gain (dB), $G_c$"); \
            plt.title("Coherent gain as a function of the number of integrated samples");

    Group:
        detection-coherent-integration
    """
    time_bandwidth = verify_arraylike(time_bandwidth, float=True, positive=True)

    g_c = db(time_bandwidth)

    return convert_output(g_c)


@export
def coherent_gain_loss(
    time: npt.ArrayLike,
    freq: npt.ArrayLike,
) -> npt.NDArray[np.float64]:
    r"""
    Computes the coherent gain loss (CGL) given a time or frequency offset.

    Arguments:
        time: The coherent integration time $T_c$ or time offset in $\Delta t$ in seconds.
        freq: The frequency offset $\Delta f$ or signal bandwidth $B_c$ in Hz.

    Returns:
        The coherent gain loss (CGL) in dB.

    Notes:
        Coherent gain loss is the reduction in SNR due to time or frequency offsets during coherent integration.
        These losses are similar to scalloping loss.

        The coherent gain loss of a signal integrated for $T_c$ seconds with a frequency offset of $\Delta f$ Hz is

        $$\text{CGL} = -10 \log_{10} \left( \text{sinc}^2 \left( T_c \Delta f \right) \right) ,$$

        where the sinc function is defined as

        $$\text{sinc}(x) = \frac{\sin(\pi x)}{\pi x} .$$

        The coherent gain loss of a signal with bandwidth $B_c$ Hz with a detection time offset of $\Delta t$ seconds
        is

        $$\text{CGL} = -10 \log_{10} \left( \text{sinc}^2 \left( \Delta t B_c \right) \right) .$$

    Examples:
        Compute the coherent gain loss for an integration time of 1 ms and a frequency offset of 235 Hz.

        .. ipython:: python

            sdr.coherent_gain_loss(1e-3, 235)

        Compute the coherent gain loss for a signal with 1 MHz of bandwidth and a detection time offset of 0.25 μs.

        .. ipython:: python

            sdr.coherent_gain_loss(0.25e-6, 1e6)

        Compute the coherent gain loss of a signal detected between two DFT bins. This is commonly referred as the DFT
        scalloping loss. Suppose the DFT is 1 ms long, then the bin spacing is 1 kHz. The worst case scalloping loss
        occurs at 1/2 bin spacing, or 500 Hz in this example. Scalloping loss of 3.9 dB for an unwindowed DFT is a
        well-known figure.

        .. ipython:: python

            t_c = 1e-3  # s
            sdr.coherent_gain_loss(t_c, 1 / t_c / 2)

        If the DFT is zero-padded to twice the length, the scalloping loss is reduced to 0.9 dB.

        .. ipython:: python

            t_c = 1e-3  # s
            sdr.coherent_gain_loss(t_c, 1 / (2 * t_c) / 2)

        Compute the coherent gain loss for an integration time of 1 ms and an array of frequency offsets.

        .. ipython:: python

            sdr.coherent_gain_loss(1e-3, [0, 100, 200, 300, 400, 500])

        Plot coherent gain loss as a function of frequency offset.

        .. ipython:: python

            f = np.linspace(0, 2e3, 1001)

            @savefig sdr_coherent_gain_loss_1.png
            plt.figure(); \
            plt.plot(f, sdr.coherent_gain_loss(0.5e-3, f), label="0.5 ms"); \
            plt.plot(f, sdr.coherent_gain_loss(1e-3, f), label="1 ms"); \
            plt.plot(f, sdr.coherent_gain_loss(2e-3, f), label="2 ms"); \
            plt.legend(title="Integration time"); \
            plt.ylim(-5, 55); \
            plt.xlabel("Frequency offset (Hz)"); \
            plt.ylabel("Coherent gain loss (dB)"); \
            plt.title("Coherent gain loss for various integration times");

        Plot coherent gain loss as a function of integration time.

        .. ipython:: python

            t = np.linspace(0, 1e-2, 1001)

            @savefig sdr_coherent_gain_loss_2.png
            plt.figure(); \
            plt.plot(t * 1e3, sdr.coherent_gain_loss(t, 100), label="100 Hz"); \
            plt.plot(t * 1e3, sdr.coherent_gain_loss(t, 200), label="200 Hz"); \
            plt.plot(t * 1e3, sdr.coherent_gain_loss(t, 400), label="400 Hz"); \
            plt.legend(title="Frequency offset"); \
            plt.ylim(-5, 55); \
            plt.xlabel("Integration time (ms)"); \
            plt.ylabel("Coherent gain loss (dB)"); \
            plt.title("Coherent gain loss for various frequency offsets");

    Group:
        detection-coherent-integration
    """
    time = verify_arraylike(time, float=True, non_negative=True)
    freq = verify_arraylike(freq, float=True)

    cgl = np.sinc(time * freq) ** 2
    cgl_db = -1 * db(cgl)

    return convert_output(cgl_db)


@export
def max_integration_time(
    cgl: npt.ArrayLike,
    freq_offset: npt.ArrayLike,
) -> npt.NDArray[np.float64]:
    r"""
    Computes the maximum integration time that produces at most the provided coherent gain loss (CGL).

    Arguments:
        cgl: The coherent gain loss (CGL) in dB.
        freq_offset: The frequency offset $\Delta f$ in Hz.

    Returns:
        The maximum integration time $T_c$ in seconds.

    Notes:
        The inverse sinc function is calculated using numerical techniques.

    Examples:
        Compute the maximum integration time that produces at most 3 dB of coherent gain loss for a frequency offset
        of 235 Hz.

        .. ipython:: python

            sdr.max_integration_time(3, 235)

        Compute the maximum integration time that produces at most 3 dB of coherent gain loss for an array of
        frequency offsets.

        .. ipython:: python

            sdr.max_integration_time(3, [0, 100, 200, 300, 400, 500])

        Plot the maximum integration time as a function of frequency offset.

        .. ipython:: python

            f = np.linspace(0, 1e3, 1001)

            @savefig sdr_max_integration_time_1.png
            plt.figure(); \
            plt.plot(f, sdr.max_integration_time(0.1, f) * 1e3, label="0.1 dB"); \
            plt.plot(f, sdr.max_integration_time(1, f) * 1e3, label="1 dB"); \
            plt.plot(f, sdr.max_integration_time(3, f) * 1e3, label="3 dB"); \
            plt.legend(); \
            plt.ylim(0, 10); \
            plt.xlabel("Frequency offset (Hz)"); \
            plt.ylabel("Maximum integration time (ms)"); \
            plt.title("Maximum integration time for various coherent gain losses");

        Plot the maximum integration time as a function of coherent gain loss.

        .. ipython:: python

            cgl = np.linspace(0, 10, 1001)

            @savefig sdr_max_integration_time_2.png
            plt.figure(); \
            plt.plot(cgl, sdr.max_integration_time(cgl, 50) * 1e3, label="50 Hz"); \
            plt.plot(cgl, sdr.max_integration_time(cgl, 100) * 1e3, label="100 Hz"); \
            plt.plot(cgl, sdr.max_integration_time(cgl, 200) * 1e3, label="200 Hz"); \
            plt.legend(); \
            plt.ylim(0, 10); \
            plt.xlabel("Coherent gain loss (dB)"); \
            plt.ylabel("Maximum integration time (ms)"); \
            plt.title("Maximum integration time for various frequency offsets");

    Group:
        detection-coherent-integration
    """
    cgl = verify_arraylike(cgl, float=True, non_negative=True)
    freq_offset = verify_arraylike(freq_offset, float=True)

    @np.vectorize
    def _calculate(cgl: float, freq_offset: float) -> float:
        if freq_offset == 0:
            return np.inf

        min_t = 0  # s
        max_t = 1 / freq_offset  # s
        t = scipy.optimize.brentq(lambda t: coherent_gain_loss(freq_offset, t) - cgl, min_t, max_t)

        return t

    t = _calculate(cgl, freq_offset)

    return convert_output(t)


@export
def max_frequency_offset(
    cgl: npt.ArrayLike,
    integration_time: npt.ArrayLike,
) -> npt.NDArray[np.float64]:
    r"""
    Computes the maximum frequency offset that produces at most the provided coherent gain loss (CGL).

    Arguments:
        cgl: The coherent gain loss (CGL) in dB.
        integration_time: The coherent integration time $T_c$ in seconds.

    Returns:
        The maximum frequency offset $\Delta f$ in Hz.

    Notes:
        The inverse sinc function is calculated using numerical techniques.

    Examples:
        Compute the maximum frequency offset that produces at most 3 dB of coherent gain loss for an integration time
        of 1 ms.

        .. ipython:: python

            sdr.max_frequency_offset(3, 1e-3)

        Compute the maximum frequency offset that produces at most 3 dB of coherent gain loss for an array of
        integration times.

        .. ipython:: python

            sdr.max_frequency_offset(3, [1e-3, 2e-3, 3e-3])

        Plot the maximum frequency offset as a function of integration time.

        .. ipython:: python

            t = np.linspace(0, 10e-3, 1001)

            @savefig sdr_max_frequency_offset_1.png
            plt.figure(); \
            plt.plot(t * 1e3, sdr.max_frequency_offset(0.1, t), label="0.1 dB"); \
            plt.plot(t * 1e3, sdr.max_frequency_offset(1, t), label="1 dB"); \
            plt.plot(t * 1e3, sdr.max_frequency_offset(3, t), label="3 dB"); \
            plt.legend(); \
            plt.ylim(0, 1e3); \
            plt.xlabel("Integration time (ms)"); \
            plt.ylabel("Maximum frequency offset (Hz)"); \
            plt.title("Maximum frequency offset for various coherent gain losses");

        Plot the maximum frequency offset as a function of coherent gain loss.

        .. ipython:: python

            cgl = np.linspace(0, 10, 1001)

            @savefig sdr_max_frequency_offset_2.png
            plt.figure(); \
            plt.plot(cgl, sdr.max_frequency_offset(cgl, 0.5e-3), label="0.5 ms"); \
            plt.plot(cgl, sdr.max_frequency_offset(cgl, 1e-3), label="1 ms"); \
            plt.plot(cgl, sdr.max_frequency_offset(cgl, 2e-3), label="2 ms"); \
            plt.legend(); \
            plt.ylim(0, 1e3); \
            plt.xlabel("Coherent gain loss (dB)"); \
            plt.ylabel("Maximum frequency offset (Hz)"); \
            plt.title("Maximum frequency offset for various integration times");

    Group:
        detection-coherent-integration
    """
    cgl = verify_arraylike(cgl, float=True, non_negative=True)
    integration_time = verify_arraylike(integration_time, float=True, non_negative=True)

    @np.vectorize
    def _calculate(cgl: float, integration_time: float) -> float:
        if integration_time == 0:
            return np.inf

        min_f = 0  # Hz
        max_f = 1 / integration_time  # Hz
        f = scipy.optimize.brentq(lambda f: coherent_gain_loss(integration_time, f) - cgl, min_f, max_f)

        return f

    f = _calculate(cgl, integration_time)

    return convert_output(f)



import pytest

import sdr


def test_stein_1():
    snr1 = -10  # dB
    snr2 = -10  # dB
    time = 1e-3  # seconds
    bandwidth = 1e6  # Hz
    sigma_tdoa = sdr.tdoa_crlb(snr1, snr2, time, bandwidth)
    assert sigma_tdoa == pytest.approx(135.04744742356585e-9)  # Stein says ~140 ns


def test_stein_2():
    snr1 = -10  # dB
    snr2 = -10  # dB
    time = 6e-3  # seconds
    bandwidth = 1e6  # Hz
    sigma_tdoa = sdr.tdoa_crlb(snr1, snr2, time, bandwidth)
    assert sigma_tdoa == pytest.approx(55.132889542179175e-9)  # Stein says ~56 ns



import numpy as np
import scipy.stats

import sdr


def test_normal():
    X = scipy.stats.norm(loc=-1, scale=0.5)
    _verify(X, 2)
    _verify(X, 3)
    _verify(X, 4)


def test_rayleigh():
    X = scipy.stats.rayleigh(scale=1)
    _verify(X, 2)
    _verify(X, 3)
    _verify(X, 4)


def test_rician():
    X = scipy.stats.rice(2)
    _verify(X, 2)
    _verify(X, 3)
    _verify(X, 4)


def _verify(X, n):
    # Numerically compute the distribution
    Y = sdr.sum_distribution(X, n)

    # Empirically compute the distribution
    y = X.rvs((100_000, n)).sum(axis=1)
    hist, bins = np.histogram(y, bins=101, density=True)
    x = bins[1:] - np.diff(bins) / 2

    if False:
        import matplotlib.pyplot as plt

        plt.figure()
        plt.plot(x, X.pdf(x), label="X")
        plt.plot(x, Y.pdf(x), label="X + ... + X")
        plt.hist(y, bins=101, cumulative=False, density=True, histtype="step", label="X + .. + X empirical")
        plt.legend()
        plt.xlabel("Random variable")
        plt.ylabel("Probability density")
        plt.title("Sum of distribution")
        plt.show()

    assert np.allclose(Y.pdf(x), hist, atol=np.max(hist) * 0.1)



"""
A module containing phase error detectors (PEDs).
"""

from __future__ import annotations

import numpy as np
import numpy.typing as npt

from .._helper import convert_output, export, verify_arraylike, verify_isinstance, verify_scalar
from .._modulation import LinearModulation


@export
class PED:
    r"""
    Implements a heuristic phase error detector (PED).

    Notes:
        The heuristic data-aided PED computes the angle between the received complex symbols $\tilde{a}[k]$
        and the known transmitted complex symbols $a[k]$.

        $$\theta_{e,DA}[k] = \angle \left( \tilde{a}[k] \cdot a^*[k] \right)$$

        The heuristic decision-directed PED computes the angle between the received complex symbols $\tilde{a}[k]$
        and the complex symbol decisions $\hat{a}[k]$.

        $$\theta_{e,DD}[k] = \angle \left( \tilde{a}[k] \cdot \hat{a}^*[k] \right)$$

    References:
        - Michael Rice, *Digital Communications: A Discrete-Time Approach*, Section 7.2.1.

    Examples:
        Compare the data-aided and decision-directed PEDs on QPSK modulation.

        .. ipython:: python

            qpsk = sdr.PSK(4)
            ped = sdr.PED()

        .. ipython:: python

            error, da_error = ped.data_aided_error(qpsk); \
            error, dd_error = ped.decision_directed_error(qpsk)

            @savefig sdr_PED_1.png
            plt.figure(); \
            plt.plot(error, da_error, label="Data-aided"); \
            plt.plot(error, dd_error, label="Decision-directed"); \
            plt.grid(True, linestyle="--"); \
            plt.legend(); \
            plt.xlabel("Phase of received symbols (radians)"); \
            plt.ylabel("Phase error (radians)"); \
            plt.title("Comparison of data-aided and decision-directed PEDs on QPSK");

        Observe that the slope of the phase error $K_p = 1$ is the same for both the data-aided and decision-directed
        PEDs. Also note that the unambiguous range of the data-aided PED is $[-\pi, \pi)$ and the
        decision-directed PED is $[-\pi/M, \pi/M)$.

    Group:
        synchronization-ped
    """

    def __init__(self) -> None:
        """
        Initializes the PED.
        """
        return

    def __call__(
        self,
        received: npt.ArrayLike,
        reference: npt.ArrayLike,
    ) -> npt.NDArray[np.float64]:
        r"""
        Detects the phase error.

        Arguments:
            received: The received complex symbols $\tilde{a}[k]$.
            reference: The reference complex symbols, either the known transmitted complex symbols $a[k]$
                or the complex symbols decisions $\hat{a}[k]$.

        Returns:
            The detected phase error $\theta_e[k]$ in radians.
        """
        received = verify_arraylike(received, complex=True, atleast_1d=True, ndim=1)
        reference = verify_arraylike(reference, complex=True, atleast_1d=True, ndim=1)

        reference = reference / np.abs(reference)  # Normalize the reference symbols
        phase_error = np.angle(received * reference.conj())

        return convert_output(phase_error)

    def data_aided_error(
        self,
        modem: LinearModulation,
        n_points: int = 1000,
    ) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
        r"""
        Simulates the average phase error of the data-aided PED using the specified modulation scheme.

        Arguments:
            modem: The linear modulation scheme.
            n_points: The number of points in the simulation.

        Returns:
            - The true phase error in radians.
            - The detected phase error in radians.
        """
        return _data_aided_error(self, modem, n_points)

    def decision_directed_error(
        self,
        modem: LinearModulation,
        n_points: int = 1000,
    ) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
        r"""
        Simulates the average phase error of the decision-directed PED using the specified modulation scheme.

        Arguments:
            modem: The linear modulation scheme.
            n_points: The number of points in the simulation.

        Returns:
            - The true phase error in radians.
            - The detected phase error in radians.
        """
        return _decision_directed_error(self, modem, n_points)

    @property
    def gain(self) -> float:
        r"""
        The gain of the phase error detector $K_p$.
        """
        return 1.0


@export
class MLPED(PED):
    r"""
    Implements a maximum-likelihood phase error detector (ML-PED).

    Notes:
        The data-aided ML-PED is computed using the received complex symbols $\tilde{a}[k]$ and the known
        transmitted complex symbols $a[k]$.

        $$\theta_{e,DA}[k] = \Im(\tilde{a}[k]) \cdot \Re(a[k]) - \Re(\tilde{a}[k]) \cdot \Im(a[k])) $$

        The data-aided ML-PED is computed using the received complex symbols $\tilde{a}[k]$ and the complex
        symbol decisions $\hat{a}[k]$.

        $$\theta_{e,DA}[k] = \Im(\tilde{a}[k]) \cdot \Re(\hat{a}[k]) - \Re(\tilde{a}[k]) \cdot \Im(\hat{a}[k]) $$

    References:
        - Michael Rice, *Digital Communications: A Discrete-Time Approach*, Section 7.2.2.

    Examples:
        Compare the data-aided and decision-directed ML-PEDs on QPSK modulation.

        .. ipython:: python

            qpsk = sdr.PSK(4)
            A_rx, A_ref = 5, 3
            ped = sdr.MLPED(A_rx, A_ref)

        .. ipython:: python

            error, da_error = ped.data_aided_error(qpsk); \
            error, dd_error = ped.decision_directed_error(qpsk)

            @savefig sdr_MLPED_1.png
            plt.figure(); \
            plt.plot(error, da_error, label="Data-aided"); \
            plt.plot(error, dd_error, label="Decision-directed"); \
            plt.grid(True, linestyle="--"); \
            plt.legend(); \
            plt.xlabel("Phase of received symbols (radians)"); \
            plt.ylabel("Phase error (radians)"); \
            plt.title("Comparison of data-aided and decision-directed ML-PEDs on QPSK");

        Observe that the slope of the phase error $K_p = A_{rx,rms} A_{ref,rms}$ is the same for both the data-aided
        and decision-directed PEDs. It's very important to observe that the gain of the ML-PED is scaled by the
        received signal amplitude $A_{rx,rms}$ and the reference signal amplitude $A_{ref,rms}$. Because of this,
        the ML-PED should only be used with automatic gain control (AGC).

        .. ipython:: python

            ped.gain
            A_rx * A_ref

        Also note that the unambiguous range of the data-aided ML-PED is $[-\pi, \pi)$ and the
        decision-directed ML-PED is $[-\pi/M, \pi/M)$.

    Group:
        synchronization-ped
    """

    def __init__(
        self,
        A_received: float = 1.0,
        A_reference: float = 1.0,
    ) -> None:
        """
        Initializes the ML-PED.

        Arguments:
            A_received: The received signal RMS amplitude $A_{rx,rms}$.
            A_reference: The reference signal RMS amplitude $A_{ref,rms}$.
        """
        self.A_received = A_received  # Uses the property setter
        self.A_reference = A_reference  # Uses the property setter

    def __call__(
        self,
        received: npt.ArrayLike,
        reference: npt.ArrayLike,
    ) -> np.ndarray:
        received = verify_arraylike(received, complex=True, atleast_1d=True, ndim=1)
        reference = verify_arraylike(reference, complex=True, atleast_1d=True, ndim=1)

        phase_error = received.imag * reference.real - received.real * reference.imag

        return convert_output(phase_error)

    def data_aided_error(
        self,
        modem: LinearModulation,
        n_points: int = 1000,
    ) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
        return _data_aided_error(self, modem, n_points, self._A_received, self._A_reference)

    def decision_directed_error(
        self,
        modem: LinearModulation,
        n_points: int = 1000,
    ) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
        return _decision_directed_error(self, modem, n_points, self._A_received, self._A_reference)

    @property
    def gain(self) -> float:
        return self.A_received * self.A_reference

    @property
    def A_received(self) -> float:
        """
        (Settable) The received signal RMS amplitude $A_{rx,rms}$.
        """
        return self._A_received

    @A_received.setter
    def A_received(self, A_received: float) -> None:
        self._A_received = verify_scalar(A_received, float=True, positive=True)

    @property
    def A_reference(self) -> float:
        """
        (Settable) The reference signal RMS amplitude $A_{ref,rms}$.
        """
        return self._A_reference

    @A_reference.setter
    def A_reference(self, A_reference: float) -> None:
        self._A_reference = verify_scalar(A_reference, float=True, positive=True)


def _data_aided_error(
    ped: PED,
    modem: LinearModulation,
    n_points: int = 1000,
    A_received: float = 1.0,
    A_reference: float = 1,
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    verify_isinstance(ped, PED)
    verify_isinstance(modem, LinearModulation)
    verify_scalar(n_points, int=True, positive=True)
    verify_scalar(A_received, float=True, positive=True)
    verify_scalar(A_reference, float=True, positive=True)

    error = np.linspace(-np.pi, np.pi, n_points)
    da_error = np.zeros(n_points, dtype=float)
    for reference in modem.symbol_map:
        received = reference * np.exp(1j * error)
        da_error += ped(A_received * received, A_reference * reference)
    da_error /= modem.order

    return error, da_error


def _decision_directed_error(
    ped: PED,
    modem: LinearModulation,
    n_points: int = 1000,
    A_received: float = 1.0,
    A_reference: float = 1,
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    verify_isinstance(ped, PED)
    verify_isinstance(modem, LinearModulation)
    verify_scalar(n_points, int=True, positive=True)
    verify_scalar(A_received, float=True, positive=True)
    verify_scalar(A_reference, float=True, positive=True)

    error = np.linspace(-np.pi, np.pi, n_points)
    dd_error = np.zeros(n_points, dtype=float)
    for reference in modem.symbol_map:
        received = reference * np.exp(1j * error)
        _, decided = modem.decide_symbols(received)
        dd_error += ped(A_received * received, A_reference * decided)
    dd_error /= modem.order

    return error, dd_error



import numpy as np
import pytest

import sdr


def test_call_exceptions():
    nco = sdr.NCO()
    with pytest.raises(ValueError):
        nco()
    with pytest.raises(ValueError):
        nco(np.array([1, 2, 3]), output="invalid")


def test_initial_phase_is_offset():
    gain = 1
    increment = 2 * np.pi / 10
    offset = np.pi / 2
    nco = sdr.NCO(gain, increment, offset)

    assert nco.gain == gain
    assert nco.increment == increment
    assert nco.offset == offset

    y = nco.step(1)
    y_truth = np.exp(1j * offset)
    assert np.allclose(y, y_truth)


def test_constant_increment():
    gain = 1
    increment = 0.1
    offset = 0
    nco = sdr.NCO(gain, increment, offset)

    assert nco.gain == gain
    assert nco.increment == increment
    assert nco.offset == offset

    N = 100
    y = nco.step(N)
    y_truth = np.exp(1j * (np.arange(N) * increment + offset))
    assert np.allclose(y, y_truth)


def test_constant_increment_constant_offset():
    gain = 1
    increment = 0.1
    offset = 0.5
    nco = sdr.NCO(gain, increment, offset)

    assert nco.gain == gain
    assert nco.increment == increment
    assert nco.offset == offset

    N = 100
    y = nco.step(N)
    y_truth = np.exp(1j * (np.arange(N) * increment + offset))
    assert np.allclose(y, y_truth)


def test_varying_phase():
    nco = sdr.NCO()

    phase = np.array([0, 1, 0, 1, 0, 1]) * np.pi
    phase = np.repeat(phase, 5)

    y = nco(phase=phase)
    y_truth = np.exp(1j * phase)
    assert np.allclose(y, y_truth)



import numpy as np
import pytest

import sdr


def test_absolute_limit():
    assert sdr.shannon_limit_ebn0(0) == pytest.approx(-1.591745389548616)


def test_vector():
    """
    These numbers were verified visually from published papers.
    """
    rate = np.array([1 / 8, 1 / 4, 1 / 3, 1 / 2, 2 / 3, 3 / 4, 7 / 8])  # Code rate, n/k
    rho = 2 * rate  # bits/2D
    ebn0 = sdr.shannon_limit_ebn0(rho)
    ebn0_truth = np.array(
        [
            -1.21002545,
            -0.8174569,
            -0.54974021,  # Verified from a paper
            0.0,
            0.56859734,
            0.85986396,
            1.30533298,
        ],
    )
    assert np.allclose(ebn0, ebn0_truth)



"""
A module containing various interleavers.
"""

from __future__ import annotations

import numpy as np
import numpy.typing as npt

from .._helper import convert_output, export, verify_arraylike, verify_condition, verify_scalar


@export
class Interleaver:
    r"""
    Implements an arbitrary interleaver.

    Examples:
        Create a length-4 interleaver with custom permutation map.

        .. ipython:: python

            pi = np.array([0, 3, 1, 2])
            interleaver = sdr.Interleaver(pi)
            interleaver.map
            interleaver.inverse_map

        Interleave and deinterleave a sequence.

        .. ipython:: python

            x = np.arange(16); x
            y = interleaver.interleave(x); y
            interleaver.deinterleave(y)

    Group:
        coding-interleavers
    """

    def __init__(self, map: npt.ArrayLike):
        r"""
        Creates an arbitrary interleaver.

        Arguments:
            map: The interleaver permutation map $\pi : i \mapsto j$, containing the values $[0, N)$.
                The $i$-th input element will be placed at the $\pi(i)$-th output position.
        """
        map = verify_arraylike(map, int=True, atleast_1d=True, ndim=1)
        verify_condition(np.unique(map).size == map.size)

        self._map = map
        self._inverse_map = np.argsort(map)

    def __len__(self) -> int:
        """
        The size of the interleaver.
        """
        return self.map.size

    def interleave(self, x: npt.ArrayLike) -> npt.NDArray:
        r"""
        Interleaves the input sequence $x[n]$.

        Arguments:
            x: The input sequence $x[n]$. Length must be a multiple of the interleaver size.

        Returns:
            The interleaved sequence $y[n]$.
        """
        x = verify_arraylike(x, ndim=1, size_multiple=len(self))

        y = np.empty_like(x)
        y.reshape((-1, len(self)))[..., self.map] = x.reshape((-1, len(self)))

        return convert_output(y)

    def deinterleave(self, y: npt.ArrayLike) -> npt.NDArray:
        r"""
        Deinterleaves the input sequence $y[n]$.

        Arguments:
            y: The input sequence $y[n]$. Length must be a multiple of the interleaver size.

        Returns:
            The deinterleaved sequence $x[n]$.
        """
        y = verify_arraylike(y, ndim=1, size_multiple=len(self))

        x = np.empty_like(y)
        x.reshape((-1, len(self)))[..., self.inverse_map] = y.reshape((-1, len(self)))

        return convert_output(x)

    @property
    def map(self) -> npt.NDArray[np.int_]:
        r"""
        The interleaver permutation map $\pi$.

        The map $\pi : i \mapsto j$ indicates that the $i$-th interleaver input will be placed at the $\pi(i)$-th
        output position.
        """
        return self._map

    @property
    def inverse_map(self) -> npt.NDArray[np.int_]:
        r"""
        The deinterleaver permutation map $\pi^{-1}$.

        The map $\pi^{-1} : j \mapsto i$ indicates that the $j$-th deinterleaver input will be placed at the
        $\pi^{-1}(j)$-th output position.
        """
        return self._inverse_map


@export
class BlockInterleaver(Interleaver):
    r"""
    Implements a block interleaver.

    Notes:
        A block interleaver feeds the input down the columns of a $R \times C$ matrix and then reads the output
        across the rows.

    Examples:
        Create a $3 \times 4$ block interleaver.

        .. ipython:: python

            interleaver = sdr.BlockInterleaver(3, 4)
            interleaver.map
            interleaver.inverse_map

        Interleave and deinterleave a sequence.

        .. ipython:: python

            x = np.arange(12); x
            y = interleaver.interleave(x); y
            interleaver.deinterleave(y)

    Group:
        coding-interleavers
    """

    def __init__(self, rows: int, cols: int):
        r"""
        Creates a $R \times C$ block interleaver.

        Arguments:
            rows: The number of rows $R$ in the interleaver. The row size determines the output separation of
                consecutive input elements.
            cols: The number of columns $C$ in the interleaver.
        """
        verify_scalar(rows, int=True, positive=True)
        verify_scalar(cols, int=True, positive=True)

        map = np.arange(rows * cols).reshape((rows, cols)).T.ravel()
        super().__init__(map)



import numpy as np

import sdr


def test_backward():
    iir = sdr.Integrator("backward")
    b_truth = [0, 1]
    a_truth = [1, -1]
    assert np.allclose(iir.b_taps, b_truth)
    assert np.allclose(iir.a_taps, a_truth)


def test_trapezoidal():
    iir = sdr.Integrator("trapezoidal")
    b_truth = [0.5, 0.5]
    a_truth = [1, -1]
    assert np.allclose(iir.b_taps, b_truth)
    assert np.allclose(iir.a_taps, a_truth)


def test_forward():
    iir = sdr.Integrator("forward")
    b_truth = [1]
    a_truth = [1, -1]
    assert np.allclose(iir.b_taps, b_truth)
    assert np.allclose(iir.a_taps, a_truth)


def test_gaussian():
    iir = sdr.Integrator("forward")
    x = sdr.gaussian(0.3, 5, 10)
    y = iir(x)
    y_truth = np.cumsum(x)
    assert np.allclose(y, y_truth)


def test_raised_cosine():
    iir = sdr.Integrator("forward")
    x = sdr.root_raised_cosine(0.1, 8, 10)
    y = iir(x)
    y_truth = np.cumsum(x)
    assert np.allclose(y, y_truth)



"""
A subpackage for various detection algorithms.
"""

from ._approximation import *
from ._coherent_integration import *
from ._correlator import *
from ._energy import *
from ._non_coherent_integration import *
from ._theory import *



import numpy as np
import pytest

import sdr


def test_scalar():
    assert sdr.coherent_gain(1) == pytest.approx(0.0)
    assert sdr.coherent_gain(2) == pytest.approx(3.010299956639813)
    assert sdr.coherent_gain(10) == pytest.approx(10.0)
    assert sdr.coherent_gain(20) == pytest.approx(13.010299956639813)


def test_vector():
    n_c = np.array([1, 2, 10, 20])
    g_c = sdr.coherent_gain(n_c)
    g_c_truth = np.array([0.0, 3.010299956639813, 10.0, 13.010299956639813])
    assert isinstance(g_c, np.ndarray)
    assert np.allclose(g_c, g_c_truth)



"""
A subpackage for various plotting functions using :obj:`matplotlib`.
"""

from ._detection import *
from ._filter import *
from ._freq_domain import *
from ._modulation import *
from ._rc_params import *
from ._spectral_estimation import *
from ._time_domain import *
from ._utility import *



"""
A module containing correlation-based detectors.
"""

from __future__ import annotations

import numpy as np
import numpy.typing as npt

from .._conversion import linear
from .._helper import convert_output, export, verify_arraylike, verify_scalar
from .._probability import Q, Qinv


@export
class ReplicaCorrelator:
    r"""
    Implements an clairvoyant replica-correlator detector.

    Notes:
        The replica-correlator detector is a clairvoyant detector that assumes perfect knowledge of the signal
        $s[n]$. The complex noise $w[n] \sim \mathcal{CN}(0, \sigma^2)$. The null and alternative hypotheses are
        given by:

        $$\mathcal{H}_0: x[n] = w[n]$$
        $$\mathcal{H}_1: x[n] = s[n] + w[n]$$

        The test statistic $T(x)$ is given by:

        $$T(x) = \mathrm{Re}\left( \sum\limits_{n=0}^{N-1} x[n]s^*[n] \right) > \gamma'$$
        $$
        T(x) \sim
        \begin{cases}
        \mathcal{N}\left(0, \sigma^2 \mathcal{E} / 2 \right) & \text{under } \mathcal{H}_0 \\
        \mathcal{N}\left(\mathcal{E}, \sigma^2 \mathcal{E} / 2 \right) & \text{under } \mathcal{H}_1 \\
        \end{cases}
        $$

        where $\mathcal{E}$ is the received energy $\mathcal{E} = \sum\limits_{n=0}^{N-1} \left| s[n] \right|^2$.

        The probability of detection $P_d$, probability of false alarm $P_{fa}$, and detection threshold
        $\gamma'$ are given by:

        $$P_d = Q\left( Q^{-1}(P_{fa}) - \sqrt{\frac{2 \mathcal{E}}{\sigma^2}} \right)$$
        $$P_{fa} = Q\left(\frac{\gamma'}{\sqrt{\sigma^2 \mathcal{E} / 2}}\right)$$
        $$\gamma' = \sqrt{\sigma^2 \mathcal{E} / 2} Q^{-1}(P_{fa})$$

    References:
        - Steven Kay, *Fundamentals of Statistical Signal Processing: Detection Theory*, Sections 4.3.2 and 13.3.1.

    Group:
        detection-detectors
    """

    # def __init__(
    #     self,
    #     N_nc: int,
    #     p_fa: float = 1e-6,
    #     streaming: bool = False,
    # ):
    #     r"""
    #     Initializes the energy detector.

    #     Arguments:
    #         N_nc: The number of samples $N_{nc}$ to non-coherently integrate.
    #         p_fa: The desired probability of false alarm $P_{fa}$.
    #     """
    #     if not isinstance(N_nc, int):
    #         raise TypeError(f"Argument 'N_nc' must be an integer, not {type(N_nc)}.")
    #     if not N_nc >= 1:
    #         raise ValueError(f"Argument 'N_nc' must be greater than or equal to 1, not {N_nc}.")
    #     self._N_nc = N_nc

    #     if not isinstance(p_fa, float):
    #         raise TypeError(f"Argument 'p_fa' must be a float, not {type(p_fa)}.")
    #     if not 0 < p_fa < 1:
    #         raise ValueError(f"Argument 'p_fa' must be in the range (0, 1), not {p_fa}.")
    #     self._p_fa = p_fa

    #     if not isinstance(streaming, bool):
    #         raise TypeError(f"Argument 'streaming' must be a bool, not {type(streaming)}.")
    #     self._streaming = streaming

    #     self._fir = FIR(np.ones(N_nc), streaming=streaming)

    @staticmethod
    def roc(
        enr: float,
        p_fa: npt.ArrayLike | None = None,
        complex: bool = True,
    ) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
        r"""
        Computes the receiver operating characteristic (ROC) curve.

        Arguments:
            enr: The received energy-to-noise ratio $\mathcal{E}/\sigma^2$ in dB.
            p_fa: The probability of false alarm $P_{fa}$. If `None`, the ROC curve is computed for
                `p_fa = np.logspace(-10, 0, 101)`.
            complex: Indicates whether the signal is complex.

        Returns:
            - The probability of false alarm $P_{fa}$.
            - The probability of detection $P_d$.

        Examples:
            .. ipython:: python

                @savefig sdr_ReplicaCorrelator_roc_1.png
                plt.figure(); \
                sdr.plot.roc(*sdr.ReplicaCorrelator.roc(-10), label="ENR = -10 dB"); \
                sdr.plot.roc(*sdr.ReplicaCorrelator.roc(-5), label="ENR = -5 dB"); \
                sdr.plot.roc(*sdr.ReplicaCorrelator.roc(0), label="ENR = 0 dB"); \
                sdr.plot.roc(*sdr.ReplicaCorrelator.roc(5), label="ENR = 5 dB"); \
                sdr.plot.roc(*sdr.ReplicaCorrelator.roc(10), label="ENR = 10 dB"); \
                sdr.plot.roc(*sdr.ReplicaCorrelator.roc(15), label="ENR = 15 dB");
        """
        verify_scalar(enr, float=True)
        if p_fa is None:
            p_fa = np.logspace(-10, 0, 101)
        else:
            p_fa = verify_arraylike(p_fa, float=True, inclusive_min=0, inclusive_max=1)

        p_d = ReplicaCorrelator.p_d(enr, p_fa, complex=complex)

        return convert_output(p_fa), convert_output(p_d)

    @staticmethod
    def p_d(
        enr: npt.ArrayLike,
        p_fa: npt.ArrayLike,
        complex: bool = True,
    ) -> npt.NDArray[np.float64]:
        r"""
        Computes the probability of detection $P_d$.

        Arguments:
            enr: The received energy-to-noise ratio $\mathcal{E}/\sigma^2$ in dB.
            p_fa: The probability of false alarm $P_{fa}$.
            complex: Indicates whether the signal is complex.

        Returns:
            The probability of detection $P_d$.

        Notes:
            For real signals:

            $$P_d = Q\left( Q^{-1}(P_{fa}) - \sqrt{\frac{\mathcal{E}}{\sigma^2}} \right)$$

            For complex signals:

            $$P_d = Q\left( Q^{-1}(P_{fa}) - \sqrt{\frac{\mathcal{E}}{\sigma^2 / 2}} \right)$$

        References:
            - Steven Kay, *Fundamentals of Statistical Signal Processing: Detection Theory*,
              Equations 4.14 and 13.9.

        Examples:
            .. ipython:: python

                enr = np.linspace(0, 20, 101)

                @savefig sdr_ReplicaCorrelator_p_d_1.png
                plt.figure(); \
                sdr.plot.p_d(enr, sdr.ReplicaCorrelator.p_d(enr, 1e-1), label="$P_{fa} = 10^{-1}$"); \
                sdr.plot.p_d(enr, sdr.ReplicaCorrelator.p_d(enr, 1e-2), label="$P_{fa} = 10^{-2}$"); \
                sdr.plot.p_d(enr, sdr.ReplicaCorrelator.p_d(enr, 1e-3), label="$P_{fa} = 10^{-3}$"); \
                sdr.plot.p_d(enr, sdr.ReplicaCorrelator.p_d(enr, 1e-4), label="$P_{fa} = 10^{-4}$"); \
                sdr.plot.p_d(enr, sdr.ReplicaCorrelator.p_d(enr, 1e-5), label="$P_{fa} = 10^{-5}$"); \
                sdr.plot.p_d(enr, sdr.ReplicaCorrelator.p_d(enr, 1e-6), label="$P_{fa} = 10^{-6}$"); \
                sdr.plot.p_d(enr, sdr.ReplicaCorrelator.p_d(enr, 1e-7), label="$P_{fa} = 10^{-7}$");
        """
        enr = verify_arraylike(enr, float=True)
        p_fa = verify_arraylike(p_fa, float=True, inclusive_min=0, inclusive_max=1)

        enr_linear = linear(enr)
        if not complex:
            d2 = enr_linear  # Deflection coefficient
        else:
            d2 = 2 * enr_linear  # Deflection coefficient
        p_d = Q(Qinv(p_fa) - np.sqrt(d2))

        return convert_output(p_d)

    @staticmethod
    def p_fa(
        threshold: npt.ArrayLike,
        energy: npt.ArrayLike,
        sigma2: npt.ArrayLike,
        complex: bool = True,
    ) -> npt.NDArray[np.float64]:
        r"""
        Computes the probability of false alarm $P_{fa}$.

        Arguments:
            threshold: The threshold $\gamma'$.
            energy: The received energy $\mathcal{E} = \sum_{i=0}^{N-1} \left| s[n] \right|^2$.
            sigma2: The noise variance $\sigma^2$.
            complex: Indicates whether the signal is complex.

        Returns:
            The probability of false alarm $P_{fa}$.

        Notes:
            For real signals:

            $$P_{fa} = Q\left( \frac{\gamma'}{\sqrt{\sigma^2 \mathcal{E}}} \right)$$

            For complex signals:

            $$P_{fa} = Q\left( \frac{\gamma'}{\sqrt{\sigma^2 \mathcal{E} / 2}} \right)$$

        References:
            - Steven Kay, *Fundamentals of Statistical Signal Processing: Detection Theory*,
              Equations 4.12 and 13.6.
        """
        threshold = verify_arraylike(threshold, float=True)
        energy = verify_arraylike(energy, float=True, non_negative=True)
        sigma2 = verify_arraylike(sigma2, float=True, non_negative=True)

        if not complex:
            p_fa = Q(threshold / np.sqrt(energy * sigma2))
        else:
            p_fa = Q(threshold / np.sqrt(energy * sigma2 / 2))

        return convert_output(p_fa)

    @staticmethod
    def threshold(
        p_fa: npt.ArrayLike,
        energy: npt.ArrayLike,
        sigma2: npt.ArrayLike,
        complex: bool = True,
    ) -> npt.NDArray[np.float64]:
        r"""
        Computes the threshold $\gamma'$.

        Arguments:
            p_fa: The probability of false alarm $P_{fa}$.
            energy: The received energy $\mathcal{E} = \sum_{i=0}^{N-1} \left| s[n] \right|^2$.
            sigma2: The noise variance $\sigma^2$.
            complex: Indicates whether the signal is complex.

        Returns:
            The threshold $\gamma'$.

        Notes:
            For real signals:

            $$\gamma' = \sqrt{\sigma^2 \mathcal{E}} Q^{-1}(P_{fa})$$

            For complex signals:

            $$\gamma' = \sqrt{\sigma^2 \mathcal{E} / 2} Q^{-1}(P_{fa})$$

        References:
            - Steven Kay, *Fundamentals of Statistical Signal Processing: Detection Theory*,
              Equations 4.12 and 13.6.
        """
        p_fa = verify_arraylike(p_fa, float=True, inclusive_min=0, inclusive_max=1)
        energy = verify_arraylike(energy, float=True, non_negative=True)
        sigma2 = verify_arraylike(sigma2, float=True, non_negative=True)

        # Compute the variance of the test statistic
        if not complex:
            gamma_prime = np.sqrt(energy * sigma2) * Qinv(p_fa)
        else:
            gamma_prime = np.sqrt(energy * sigma2 / 2) * Qinv(p_fa)

        return convert_output(gamma_prime)

    # def test_statistic(self, x: npt.ArrayLike) -> npt.NDArray[np.float64]:
    #     """
    #     Computes the test statistic $T(x)$.

    #     Arguments:
    #         x: The received signal $x[n]$.

    #     Returns:
    #         The test statistic $T(x)$.
    #     """
    #     x = np.asarray(x)

    #     if self.streaming:
    #         return self._fir(np.abs(x) ** 2)

    #     return self._fir(np.abs(x) ** 2, mode="same")

    # def detect(self, x: npt.ArrayLike, sigma2: float) -> npt.NDArray[np.float64]:
    #     r"""
    #     Detects the presence of a signal.

    #     Arguments:
    #         x: The received signal $x[n]$.
    #         sigma2: The noise variance $\sigma^2$.

    #     Returns:
    #         The decision statistic $d$.
    #     """
    #     T = self.test_statistic(x)
    #     gamma = self.threshold(self.N_nc, self.desired_p_fa, sigma2)
    #     return T >= gamma

    # @property
    # def N_nc(self) -> int:
    #     """
    #     The number of samples $N_{nc}$ to non-coherently integrate.
    #     """
    #     return self._N_nc

    # @property
    # def desired_p_fa(self) -> float:
    #     """
    #     The desired probability of false alarm $P_{fa}$.
    #     """
    #     return self._p_fa

    # @property
    # def streaming(self) -> bool:
    #     """
    #     Indicates whether the detector is in streaming mode.
    #     """
    #     return self._streaming



import pytest

import sdr


def test_stein_1():
    # Modified for FOA from Stein's example
    snr1 = -10  # dB
    time = 1e-3  # seconds
    bandwidth = 1e6  # Hz
    sigma_foa = sdr.foa_crlb(snr1, time, bandwidth)
    assert sigma_foa == pytest.approx(38.9848400616838)


def test_stein_2():
    # Modified for FOA from Stein's example
    snr1 = -10  # dB
    time = 6e-3  # seconds
    bandwidth = 1e6  # Hz
    sigma_foa = sdr.foa_crlb(snr1, time, bandwidth)
    assert sigma_foa == pytest.approx(2.652582384864922)



"""
A module containing various distance measurement functions.
"""

from __future__ import annotations

import numpy as np
import numpy.typing as npt

from .._helper import convert_output, export, verify_arraylike


@export
def euclidean(
    x: npt.ArrayLike,
    y: npt.ArrayLike,
    axis: int | tuple[int, ...] | None = None,
) -> npt.NDArray[np.float64]:
    r"""
    Measures the Euclidean distance between two signals $x[n]$ and $y[n]$.

    $$d = \sqrt{\sum_{n=0}^{N-1} \left| x[n] - y[n] \right|^2}$$

    Arguments:
        x: The time-domain signal $x[n]$.
        y: The time-domain signal $y[n]$.
        axis: Axis or axes along which to compute the distance. The default is `None`, which computes the distance
            across the entire array.

    Returns:
        The Euclidean distance between $x[n]$ and $y[n]$.

    Group:
        measurement-distance
    """
    x = verify_arraylike(x, complex=True)
    y = verify_arraylike(y, complex=True)

    d = np.sqrt(np.sum(np.abs(x - y) ** 2, axis=axis))

    return convert_output(d)


@export
def hamming(
    x: npt.NDArray[np.int_],
    y: npt.NDArray[np.int_],
    axis: int | tuple[int, ...] | None = None,
) -> npt.NDArray[np.int_]:
    r"""
    Measures the Hamming distance between two signals $x[n]$ and $y[n]$.

    $$d = \sum_{n=0}^{N-1} x[n] \oplus y[n]$$

    Arguments:
        x: The time-domain signal $x[n]$.
        y: The time-domain signal $y[n]$.
        axis: Axis or axes along which to compute the distance. The default is `None`, which computes the distance
            across the entire array.

    Returns:
        The Hamming distance between $x[n]$ and $y[n]$.

    Group:
        measurement-distance
    """
    x = verify_arraylike(x, int=True)
    y = verify_arraylike(y, int=True)

    d = np.sum(np.bitwise_xor(x, y), axis=axis)

    return convert_output(d)
